import unittest
from unittest import mock

import pykka

from mopidy import core, mixer
from mopidy.internal.models import MixerState

from tests import dummy_mixer


class CoreMixerTest(unittest.TestCase):
    def setUp(self):  # noqa: N802
        self.mixer = mock.Mock(spec=mixer.Mixer)
        self.core = core.Core(mixer=self.mixer, backends=[])

    def test_get_volume(self):
        self.mixer.get_volume.return_value.get.return_value = 30

        assert self.core.mixer.get_volume() == 30
        self.mixer.get_volume.assert_called_once_with()

    def test_set_volume(self):
        self.mixer.set_volume.return_value.get.return_value = True
        self.core.mixer.set_volume(30)

        self.mixer.set_volume.assert_called_once_with(30)

    def test_get_mute(self):
        self.mixer.get_mute.return_value.get.return_value = True

        assert self.core.mixer.get_mute() is True
        self.mixer.get_mute.assert_called_once_with()

    def test_set_mute(self):
        self.mixer.set_mute.return_value.get.return_value = True
        self.core.mixer.set_mute(True)

        self.mixer.set_mute.assert_called_once_with(True)


class CoreNoneMixerTest(unittest.TestCase):
    def setUp(self):  # noqa: N802
        self.core = core.Core(mixer=None, backends=[])

    def test_get_volume_return_none_because_it_is_unknown(self):
        assert self.core.mixer.get_volume() is None

    def test_set_volume_return_false_because_it_failed(self):
        assert self.core.mixer.set_volume(30) is False

    def test_get_mute_return_none_because_it_is_unknown(self):
        assert self.core.mixer.get_mute() is None

    def test_set_mute_return_false_because_it_failed(self):
        assert self.core.mixer.set_mute(True) is False


@mock.patch.object(mixer.MixerListener, "send")
class CoreMixerListenerTest(unittest.TestCase):
    def setUp(self):  # noqa: N802
        self.mixer = dummy_mixer.create_proxy()
        self.core = core.Core(mixer=self.mixer, backends=[])

    def tearDown(self):  # noqa: N802
        pykka.ActorRegistry.stop_all()

    def test_forwards_mixer_volume_changed_event_to_frontends(self, send):
        assert self.core.mixer.set_volume(volume=60) is True
        assert send.call_args[0][0] == "volume_changed"
        assert send.call_args[1]["volume"] == 60

    def test_forwards_mixer_mute_changed_event_to_frontends(self, send):
        self.core.mixer.set_mute(mute=True)

        assert send.call_args[0][0] == "mute_changed"
        assert send.call_args[1]["mute"] is True


@mock.patch.object(mixer.MixerListener, "send")
class CoreNoneMixerListenerTest(unittest.TestCase):
    def setUp(self):  # noqa: N802
        self.core = core.Core(mixer=None, backends=[])

    def test_forwards_mixer_volume_changed_event_to_frontends(self, send):
        assert self.core.mixer.set_volume(volume=60) is False
        assert send.call_count == 0

    def test_forwards_mixer_mute_changed_event_to_frontends(self, send):
        self.core.mixer.set_mute(mute=True)
        assert send.call_count == 0


class MockBackendCoreMixerBase(unittest.TestCase):
    def setUp(self):  # noqa: N802
        self.mixer = mock.Mock()
        self.mixer.actor_ref.actor_class.__name__ = "DummyMixer"
        self.core = core.Core(mixer=self.mixer, backends=[])


class GetVolumeBadBackendTest(MockBackendCoreMixerBase):
    def test_backend_raises_exception(self):
        self.mixer.get_volume.return_value.get.side_effect = Exception
        assert self.core.mixer.get_volume() is None

    def test_backend_returns_too_small_value(self):
        self.mixer.get_volume.return_value.get.return_value = -1
        assert self.core.mixer.get_volume() is None

    def test_backend_returns_too_large_value(self):
        self.mixer.get_volume.return_value.get.return_value = 1000
        assert self.core.mixer.get_volume() is None

    def test_backend_returns_wrong_type(self):
        self.mixer.get_volume.return_value.get.return_value = "12"
        assert self.core.mixer.get_volume() is None


class SetVolumeBadBackendTest(MockBackendCoreMixerBase):
    def test_backend_raises_exception(self):
        self.mixer.set_volume.return_value.get.side_effect = Exception
        assert not self.core.mixer.set_volume(30)

    def test_backend_returns_wrong_type(self):
        self.mixer.set_volume.return_value.get.return_value = "done"
        assert not self.core.mixer.set_volume(30)


class GetMuteBadBackendTest(MockBackendCoreMixerBase):
    def test_backend_raises_exception(self):
        self.mixer.get_mute.return_value.get.side_effect = Exception
        assert self.core.mixer.get_mute() is None

    def test_backend_returns_wrong_type(self):
        self.mixer.get_mute.return_value.get.return_value = "12"
        assert self.core.mixer.get_mute() is None


class SetMuteBadBackendTest(MockBackendCoreMixerBase):
    def test_backend_raises_exception(self):
        self.mixer.set_mute.return_value.get.side_effect = Exception
        assert not self.core.mixer.set_mute(True)

    def test_backend_returns_wrong_type(self):
        self.mixer.set_mute.return_value.get.return_value = "done"
        assert not self.core.mixer.set_mute(True)


class CoreMixerSaveLoadStateTest(unittest.TestCase):
    def setUp(self):  # noqa: N802
        self.mixer = dummy_mixer.create_proxy()
        self.core = core.Core(mixer=self.mixer, backends=[])

    def test_save_mute(self):
        volume = 32
        mute = False
        target = MixerState(volume=volume, mute=mute)
        self.core.mixer.set_volume(volume)
        self.core.mixer.set_mute(mute)
        value = self.core.mixer._save_state()
        assert target == value

    def test_save_unmute(self):
        volume = 33
        mute = True
        target = MixerState(volume=volume, mute=mute)
        self.core.mixer.set_volume(volume)
        self.core.mixer.set_mute(mute)
        value = self.core.mixer._save_state()
        assert target == value

    def test_load(self):
        self.core.mixer.set_volume(11)
        volume = 45
        target = MixerState(volume=volume)
        coverage = ["mixer"]
        self.core.mixer._load_state(target, coverage)
        assert volume == self.core.mixer.get_volume()

    def test_load_not_covered(self):
        self.core.mixer.set_volume(21)
        self.core.mixer.set_mute(True)
        target = MixerState(volume=56, mute=False)
        coverage = ["other"]
        self.core.mixer._load_state(target, coverage)
        assert 21 == self.core.mixer.get_volume()
        assert self.core.mixer.get_mute() is True

    def test_load_mute_on(self):
        self.core.mixer.set_mute(False)
        assert self.core.mixer.get_mute() is False
        target = MixerState(mute=True)
        coverage = ["mixer"]
        self.core.mixer._load_state(target, coverage)
        assert self.core.mixer.get_mute() is True

    def test_load_mute_off(self):
        self.core.mixer.set_mute(True)
        assert self.core.mixer.get_mute() is True
        target = MixerState(mute=False)
        coverage = ["mixer"]
        self.core.mixer._load_state(target, coverage)
        assert self.core.mixer.get_mute() is False

    def test_load_invalid_type(self):
        with self.assertRaises(TypeError):
            self.core.mixer._load_state(11, None)

    def test_load_none(self):
        self.core.mixer._load_state(None, None)



import unittest

from mopidy.config import validators


class ValidateChoiceTest(unittest.TestCase):
    def test_no_choices_passes(self):
        validators.validate_choice("foo", None)

    def test_valid_value_passes(self):
        validators.validate_choice("foo", ["foo", "bar", "baz"])
        validators.validate_choice(1, [1, 2, 3])

    def test_empty_choices_fails(self):
        self.assertRaises(ValueError, validators.validate_choice, "foo", [])

    def test_invalid_value_fails(self):
        words = ["foo", "bar", "baz"]
        self.assertRaises(
            ValueError, validators.validate_choice, "foobar", words
        )
        self.assertRaises(ValueError, validators.validate_choice, 5, [1, 2, 3])


class ValidateMinimumTest(unittest.TestCase):
    def test_no_minimum_passes(self):
        validators.validate_minimum(10, None)

    def test_valid_value_passes(self):
        validators.validate_minimum(10, 5)

    def test_to_small_value_fails(self):
        self.assertRaises(ValueError, validators.validate_minimum, 10, 20)

    def test_to_small_value_fails_with_zero_as_minimum(self):
        self.assertRaises(ValueError, validators.validate_minimum, -1, 0)


class ValidateMaximumTest(unittest.TestCase):
    def test_no_maximum_passes(self):
        validators.validate_maximum(5, None)

    def test_valid_value_passes(self):
        validators.validate_maximum(5, 10)

    def test_to_large_value_fails(self):
        self.assertRaises(ValueError, validators.validate_maximum, 10, 5)

    def test_to_large_value_fails_with_zero_as_maximum(self):
        self.assertRaises(ValueError, validators.validate_maximum, 5, 0)


class ValidateRequiredTest(unittest.TestCase):
    def test_passes_when_false(self):
        validators.validate_required("foo", False)
        validators.validate_required("", False)
        validators.validate_required("  ", False)
        validators.validate_required([], False)

    def test_passes_when_required_and_set(self):
        validators.validate_required("foo", True)
        validators.validate_required(" foo ", True)
        validators.validate_required([1], True)

    def test_blocks_when_required_and_emtpy(self):
        self.assertRaises(ValueError, validators.validate_required, "", True)
        self.assertRaises(ValueError, validators.validate_required, [], True)



import configparser
import os
import pathlib


def get_dirs():
    """Returns a dict of all the known XDG Base Directories for the current user.

    The keys ``XDG_CACHE_DIR``, ``XDG_CONFIG_DIR``, and ``XDG_DATA_DIR`` is
    always available.

    Additional keys, like ``XDG_MUSIC_DIR``, may be available if the
    ``$XDG_CONFIG_DIR/user-dirs.dirs`` file exists and is parseable.

    See http://standards.freedesktop.org/basedir-spec/basedir-spec-latest.html
    for the XDG Base Directory specification.
    """

    dirs = {
        "XDG_CACHE_DIR": pathlib.Path(
            os.getenv("XDG_CACHE_HOME", "~/.cache")
        ).expanduser(),
        "XDG_CONFIG_DIR": pathlib.Path(
            os.getenv("XDG_CONFIG_HOME", "~/.config")
        ).expanduser(),
        "XDG_DATA_DIR": pathlib.Path(
            os.getenv("XDG_DATA_HOME", "~/.local/share")
        ).expanduser(),
    }

    dirs.update(_get_user_dirs(dirs["XDG_CONFIG_DIR"]))

    return dirs


def _get_user_dirs(xdg_config_dir):
    """Returns a dict of XDG dirs read from
    ``$XDG_CONFIG_HOME/user-dirs.dirs``.

    This is used at import time for most users of :mod:`mopidy`. By rolling our
    own implementation instead of using :meth:`glib.get_user_special_dir` we
    make it possible for many extensions to run their test suites, which are
    importing parts of :mod:`mopidy`, in a virtualenv with global site-packages
    disabled, and thus no :mod:`glib` available.
    """

    dirs_file = xdg_config_dir / "user-dirs.dirs"

    if not dirs_file.exists():
        return {}

    data = dirs_file.read_bytes()
    data = b"[XDG_USER_DIRS]\n" + data
    data = data.replace(b"$HOME", bytes(pathlib.Path.home()))
    data = data.replace(b'"', b"")

    config = configparser.RawConfigParser()
    config.read_string(data.decode())

    result = {}
    for k, v in config.items("XDG_USER_DIRS"):
        if v is None:
            continue
        if isinstance(k, bytes):
            k = k.decode()
        result[k.upper()] = pathlib.Path(v).resolve()

    return result



import pathlib
import shutil
import tempfile
import unittest

import pytest

from mopidy.internal import path
from mopidy.internal.gi import GLib


class GetOrCreateDirTest(unittest.TestCase):
    def setUp(self):  # noqa: N802
        self.parent = pathlib.Path(tempfile.mkdtemp())

    def tearDown(self):  # noqa: N802
        if self.parent.is_dir():
            shutil.rmtree(str(self.parent))

    def test_creating_dir(self):
        dir_path = self.parent / "test"
        assert not dir_path.exists()

        created = path.get_or_create_dir(str(dir_path))

        assert dir_path.is_dir()
        assert created == dir_path

    def test_creating_nested_dirs(self):
        level2_dir = self.parent / "test"
        level3_dir = self.parent / "test" / "test"
        assert not level2_dir.exists()
        assert not level3_dir.exists()

        created = path.get_or_create_dir(str(level3_dir))

        assert level2_dir.is_dir()
        assert level3_dir.is_dir()
        assert created == level3_dir

    def test_creating_existing_dir(self):
        created = path.get_or_create_dir(str(self.parent))

        assert self.parent.is_dir()
        assert created == self.parent

    def test_create_dir_with_name_of_existing_file_throws_oserror(self):
        conflicting_file = self.parent / "test"
        conflicting_file.touch()
        dir_path = self.parent / "test"

        with pytest.raises(OSError):
            path.get_or_create_dir(str(dir_path))

    def test_create_dir_with_none(self):
        with pytest.raises(TypeError):
            path.get_or_create_dir(None)


class GetOrCreateFileTest(unittest.TestCase):
    def setUp(self):  # noqa: N802
        self.parent = pathlib.Path(tempfile.mkdtemp())

    def tearDown(self):  # noqa: N802
        if self.parent.is_dir():
            shutil.rmtree(str(self.parent))

    def test_creating_file(self):
        file_path = self.parent / "test"
        assert not file_path.exists()

        created = path.get_or_create_file(str(file_path))

        assert file_path.is_file()
        assert created == file_path

    def test_creating_nested_file(self):
        level2_dir = self.parent / "test"
        file_path = self.parent / "test" / "test"
        assert not level2_dir.exists()
        assert not file_path.exists()

        created = path.get_or_create_file(str(file_path))

        assert level2_dir.is_dir()
        assert file_path.is_file()
        assert created == file_path

    def test_creating_existing_file(self):
        file_path = self.parent / "test"
        path.get_or_create_file(str(file_path))

        created = path.get_or_create_file(str(file_path))

        assert file_path.is_file()
        assert created == file_path

    def test_create_file_with_name_of_existing_dir_throws_error(self):
        with pytest.raises(OSError):
            path.get_or_create_file(self.parent)

    def test_create_file_with_none_filename_throws_type_error(self):
        with pytest.raises(TypeError):
            path.get_or_create_file(None)

    def test_create_dir_without_mkdir(self):
        file_path = self.parent / "foo" / "bar"

        with pytest.raises(OSError):
            path.get_or_create_file(file_path, mkdir=False)

    def test_create_dir_with_bytes_content(self):
        file_path = self.parent / "test"

        created = path.get_or_create_file(str(file_path), content=b"foobar")

        assert created.read_bytes() == b"foobar"

    def test_create_dir_with_unicode_content(self):
        file_path = self.parent / "test"

        created = path.get_or_create_file(str(file_path), content="foobaræøå")
        assert created.read_bytes() == b"foobar\xc3\xa6\xc3\xb8\xc3\xa5"


class GetUnixSocketPathTest(unittest.TestCase):
    def test_correctly_matched_socket_path(self):
        assert (
            path.get_unix_socket_path("unix:/tmp/mopidy.socket")
            == "/tmp/mopidy.socket"
        )

    def test_correctly_no_match_socket_path(self):
        assert path.get_unix_socket_path("127.0.0.1") is None


class PathToFileURITest(unittest.TestCase):
    def test_simple_path(self):
        result = path.path_to_uri("/etc/fstab")

        assert result == "file:///etc/fstab"

    def test_space_in_path(self):
        result = path.path_to_uri("/tmp/test this")

        assert result == "file:///tmp/test%20this"

    def test_unicode_in_path(self):
        result = path.path_to_uri("/tmp/æøå")

        assert result == "file:///tmp/%C3%A6%C3%B8%C3%A5"


class UriToPathTest(unittest.TestCase):
    def test_simple_uri(self):
        result = path.uri_to_path("file:///etc/fstab")

        assert result == pathlib.Path("/etc/fstab")

    def test_space_in_uri(self):
        result = path.uri_to_path("file:///tmp/test%20this")

        assert result == pathlib.Path("/tmp/test this")

    def test_unicode_in_uri(self):
        result = path.uri_to_path("file:///tmp/%C3%A6%C3%B8%C3%A5")

        assert result == pathlib.Path("/tmp/æøå")

    def test_latin1_in_uri(self):
        result = path.uri_to_path("file:///tmp/%E6%F8%E5")

        assert bytes(result) == b"/tmp/\xe6\xf8\xe5"


class ExpandPathTest(unittest.TestCase):
    def test_empty_path(self):
        result = path.expand_path("")

        assert result == pathlib.Path(".").resolve()

    def test_absolute_path(self):
        result = path.expand_path("/tmp/foo")

        assert result == pathlib.Path("/tmp/foo")

    def test_home_dir_expansion(self):
        result = path.expand_path("~/foo")

        assert result == pathlib.Path("~/foo").expanduser()

    def test_abspath(self):
        result = path.expand_path("./foo")

        assert result == pathlib.Path("./foo").resolve()

    def test_xdg_subsititution(self):
        expected = GLib.get_user_data_dir() + "/foo"
        result = path.expand_path("$XDG_DATA_DIR/foo")

        assert str(result) == expected

    def test_xdg_subsititution_unknown(self):
        result = path.expand_path("/tmp/$XDG_INVALID_DIR/foo")

        assert result is None

    def test_invalid_utf8_bytes(self):
        result = path.expand_path(b"ab\xc3\x12")

        assert result == pathlib.Path("ab\udcc3\x12").resolve()


class TestIsPathInsideBaseDir:
    def test_when_inside(self):
        assert path.is_path_inside_base_dir("/æ/øå", "/æ")

    def test_when_outside(self):
        assert not path.is_path_inside_base_dir("/æ/øå", "/ø")

    def test_byte_inside_str_does_not_fail(self):
        assert path.is_path_inside_base_dir("/æ/øå".encode(), "/æ")

    def test_str_inside_byte_does_not_fail(self):
        assert path.is_path_inside_base_dir("/æ/øå", "/æ".encode())

    def test_str_inside_str_fails_does_not_fail(self):
        assert path.is_path_inside_base_dir("/æ/øå", "/æ")

    def test_bytes_inside_bytes_fails_does_not_fail(self):
        assert path.is_path_inside_base_dir("/æ/øå".encode(), "/æ".encode())



from unittest import mock

import pytest

from mopidy.internal import path
from mopidy.models import Track
from mopidy.stream import actor

from tests import path_to_data_dir


@pytest.fixture
def config():
    return {
        "proxy": {},
        "stream": {
            "timeout": 1000,
            "metadata_blacklist": [],
            "protocols": ["file"],
        },
        "file": {"enabled": False},
    }


@pytest.fixture
def audio():
    return mock.Mock()


@pytest.fixture
def track_uri():
    return path.path_to_uri(path_to_data_dir("song1.wav"))


def test_lookup_ignores_unknown_scheme(audio, config):
    backend = actor.StreamBackend(audio=audio, config=config)
    assert backend.library.lookup("http://example.com") == []


def test_lookup_respects_blacklist(audio, config, track_uri):
    config["stream"]["metadata_blacklist"].append(track_uri)
    backend = actor.StreamBackend(audio=audio, config=config)

    assert backend.library.lookup(track_uri) == [Track(uri=track_uri)]


def test_lookup_respects_blacklist_globbing(audio, config, track_uri):
    blacklist_glob = path.path_to_uri(path_to_data_dir("")) + "*"
    config["stream"]["metadata_blacklist"].append(blacklist_glob)
    backend = actor.StreamBackend(audio=audio, config=config)

    assert backend.library.lookup(track_uri) == [Track(uri=track_uri)]


def test_lookup_converts_uri_metadata_to_track(audio, config, track_uri):
    backend = actor.StreamBackend(audio=audio, config=config)

    result = backend.library.lookup(track_uri)

    assert len(result) == 1
    track = result[0]
    assert track.uri == track_uri
    assert track.length == 4406



import unittest
from unittest import mock

from mopidy import backend, core
from mopidy.models import Image, Ref, SearchResult, Track


class BaseCoreLibraryTest(unittest.TestCase):
    def setUp(self):  # noqa: N802
        dummy1_root = Ref.directory(uri="dummy1:directory", name="dummy1")
        self.backend1 = mock.Mock()
        self.backend1.uri_schemes.get.return_value = ["dummy1"]
        self.backend1.actor_ref.actor_class.__name__ = "DummyBackend1"
        self.library1 = mock.Mock(spec=backend.LibraryProvider)
        self.library1.get_images.return_value.get.return_value = {}
        self.library1.root_directory.get.return_value = dummy1_root
        self.backend1.library = self.library1
        self.backend1.has_playlists.return_value.get.return_value = False

        dummy2_root = Ref.directory(uri="dummy2:directory", name="dummy2")
        self.backend2 = mock.Mock()
        self.backend2.uri_schemes.get.return_value = ["dummy2", "du2"]
        self.backend2.actor_ref.actor_class.__name__ = "DummyBackend2"
        self.library2 = mock.Mock(spec=backend.LibraryProvider)
        self.library2.get_images.return_value.get.return_value = {}
        self.library2.root_directory.get.return_value = dummy2_root
        self.backend2.library = self.library2
        self.backend2.has_playlists.return_value.get.return_value = False

        # A backend without the optional library provider
        self.backend3 = mock.Mock()
        self.backend3.uri_schemes.get.return_value = ["dummy3"]
        self.backend3.actor_ref.actor_class.__name__ = "DummyBackend3"
        self.backend3.has_library.return_value.get.return_value = False
        self.backend3.has_library_browse.return_value.get.return_value = False

        self.core = core.Core(
            mixer=None, backends=[self.backend1, self.backend2, self.backend3]
        )


# TODO: split by method
class CoreLibraryTest(BaseCoreLibraryTest):
    def test_get_images_returns_empty_dict_for_no_uris(self):
        assert {} == self.core.library.get_images([])

    def test_get_images_returns_empty_result_for_unknown_uri(self):
        result = self.core.library.get_images(["dummy4:track"])
        assert {"dummy4:track": ()} == result

    def test_get_images_returns_empty_result_for_library_less_uri(self):
        result = self.core.library.get_images(["dummy3:track"])
        assert {"dummy3:track": ()} == result

    def test_get_images_maps_uri_to_backend(self):
        self.core.library.get_images(["dummy1:track"])
        self.library1.get_images.assert_called_once_with(["dummy1:track"])
        self.library2.get_images.assert_not_called()

    def test_get_images_maps_uri_to_backends(self):
        self.core.library.get_images(["dummy1:track", "dummy2:track"])
        self.library1.get_images.assert_called_once_with(["dummy1:track"])
        self.library2.get_images.assert_called_once_with(["dummy2:track"])

    def test_get_images_returns_images(self):
        self.library1.get_images.return_value.get.return_value = {
            "dummy1:track": [Image(uri="uri")]
        }

        result = self.core.library.get_images(["dummy1:track"])
        assert {"dummy1:track": (Image(uri="uri"),)} == result

    def test_get_images_merges_results(self):
        self.library1.get_images.return_value.get.return_value = {
            "dummy1:track": [Image(uri="uri1")]
        }
        self.library2.get_images.return_value.get.return_value = {
            "dummy2:track": [Image(uri="uri2")]
        }

        result = self.core.library.get_images(
            ["dummy1:track", "dummy2:track", "dummy3:track", "dummy4:track"]
        )
        expected = {
            "dummy1:track": (Image(uri="uri1"),),
            "dummy2:track": (Image(uri="uri2"),),
            "dummy3:track": (),
            "dummy4:track": (),
        }
        assert expected == result

    def test_browse_root_returns_dir_ref_for_each_lib_with_root_dir_name(self):
        result = self.core.library.browse(None)

        assert result == [
            Ref.directory(uri="dummy1:directory", name="dummy1"),
            Ref.directory(uri="dummy2:directory", name="dummy2"),
        ]
        assert not self.library1.browse.called
        assert not self.library2.browse.called
        assert not self.backend3.library.browse.called

    def test_browse_empty_string_returns_nothing(self):
        result = self.core.library.browse("")

        assert result == []
        assert not self.library1.browse.called
        assert not self.library2.browse.called

    def test_browse_dummy1_selects_dummy1_backend(self):
        self.library1.browse.return_value.get.return_value = [
            Ref.directory(uri="dummy1:directory:/foo/bar", name="bar"),
            Ref.track(uri="dummy1:track:/foo/baz.mp3", name="Baz"),
        ]

        self.core.library.browse("dummy1:directory:/foo")

        assert self.library1.browse.call_count == 1
        assert self.library2.browse.call_count == 0
        self.library1.browse.assert_called_with("dummy1:directory:/foo")

    def test_browse_dummy2_selects_dummy2_backend(self):
        self.library2.browse.return_value.get.return_value = [
            Ref.directory(uri="dummy2:directory:/bar/baz", name="quux"),
            Ref.track(uri="dummy2:track:/bar/foo.mp3", name="Baz"),
        ]

        self.core.library.browse("dummy2:directory:/bar")

        assert self.library1.browse.call_count == 0
        assert self.library2.browse.call_count == 1
        self.library2.browse.assert_called_with("dummy2:directory:/bar")

    def test_browse_dummy3_returns_nothing(self):
        result = self.core.library.browse("dummy3:test")

        assert result == []
        assert self.library1.browse.call_count == 0
        assert self.library2.browse.call_count == 0

    def test_browse_dir_returns_subdirs_and_tracks(self):
        self.library1.browse.return_value.get.return_value = [
            Ref.directory(uri="dummy1:directory:/foo/bar", name="Bar"),
            Ref.track(uri="dummy1:track:/foo/baz.mp3", name="Baz"),
        ]

        result = self.core.library.browse("dummy1:directory:/foo")
        assert result == [
            Ref.directory(uri="dummy1:directory:/foo/bar", name="Bar"),
            Ref.track(uri="dummy1:track:/foo/baz.mp3", name="Baz"),
        ]

    def test_lookup_returns_empty_dict_for_no_uris(self):
        assert {} == self.core.library.lookup(uris=[])

    def test_lookup_can_handle_uris(self):
        track1 = Track(uri="dummy1:a", name="abc")
        track2 = Track(uri="dummy2:a", name="def")

        self.library1.lookup().get.return_value = [track1]
        self.library2.lookup().get.return_value = [track2]

        result = self.core.library.lookup(uris=["dummy1:a", "dummy2:a"])
        assert result == {
            "dummy2:a": [track2],
            "dummy1:a": [track1],
        }

    def test_lookup_uris_returns_empty_list_for_dummy3_track(self):
        result = self.core.library.lookup(uris=["dummy3:a"])

        assert result == {
            "dummy3:a": [],
        }
        assert not self.library1.lookup.called
        assert not self.library2.lookup.called

    def test_lookup_ignores_tracks_without_uri_set(self):
        track1 = Track(uri="dummy1:a", name="abc")
        track2 = Track()

        self.library1.lookup().get.return_value = [track1, track2]

        result = self.core.library.lookup(uris=["dummy1:a"])
        assert result == {
            "dummy1:a": [track1],
        }

    def test_refresh_with_uri_selects_dummy1_backend(self):
        self.core.library.refresh("dummy1:a")

        self.library1.refresh.assert_called_once_with("dummy1:a")
        assert not self.library2.refresh.called

    def test_refresh_with_uri_selects_dummy2_backend(self):
        self.core.library.refresh("dummy2:a")

        assert not self.library1.refresh.called
        self.library2.refresh.assert_called_once_with("dummy2:a")

    def test_refresh_with_uri_fails_silently_for_dummy3_uri(self):
        self.core.library.refresh("dummy3:a")

        assert not self.library1.refresh.called
        assert not self.library2.refresh.called

    def test_refresh_without_uri_calls_all_backends(self):
        self.core.library.refresh()

        self.library1.refresh.return_value.get.assert_called_once_with()
        self.library2.refresh.return_value.get.assert_called_once_with()

    def test_search_combines_results_from_all_backends(self):
        track1 = Track(uri="dummy1:a")
        track2 = Track(uri="dummy2:a")
        result1 = SearchResult(tracks=[track1])
        result2 = SearchResult(tracks=[track2])

        self.library1.search.return_value.get.return_value = result1
        self.library2.search.return_value.get.return_value = result2

        result = self.core.library.search({"any": ["a"]})

        assert result1 in result
        assert result2 in result
        self.library1.search.assert_called_once_with(
            query={"any": ["a"]}, uris=None, exact=False
        )
        self.library2.search.assert_called_once_with(
            query={"any": ["a"]}, uris=None, exact=False
        )

    def test_search_with_uris_selects_dummy1_backend(self):
        self.core.library.search(
            query={"any": ["a"]}, uris=["dummy1:", "dummy1:foo", "dummy3:"]
        )

        self.library1.search.assert_called_once_with(
            query={"any": ["a"]}, uris=["dummy1:", "dummy1:foo"], exact=False
        )
        assert not self.library2.search.called

    def test_search_with_uris_selects_both_backends(self):
        self.core.library.search(
            query={"any": ["a"]}, uris=["dummy1:", "dummy1:foo", "dummy2:"]
        )

        self.library1.search.assert_called_once_with(
            query={"any": ["a"]}, uris=["dummy1:", "dummy1:foo"], exact=False
        )
        self.library2.search.assert_called_once_with(
            query={"any": ["a"]}, uris=["dummy2:"], exact=False
        )

    def test_search_filters_out_none(self):
        track1 = Track(uri="dummy1:a")
        result1 = SearchResult(tracks=[track1])

        self.library1.search.return_value.get.return_value = result1
        self.library2.search.return_value.get.return_value = None

        result = self.core.library.search({"any": ["a"]})

        assert result1 in result
        assert None not in result
        self.library1.search.assert_called_once_with(
            query={"any": ["a"]}, uris=None, exact=False
        )
        self.library2.search.assert_called_once_with(
            query={"any": ["a"]}, uris=None, exact=False
        )

    def test_search_accepts_query_dict_instead_of_kwargs(self):
        track1 = Track(uri="dummy1:a")
        track2 = Track(uri="dummy2:a")
        result1 = SearchResult(tracks=[track1])
        result2 = SearchResult(tracks=[track2])

        self.library1.search.return_value.get.return_value = result1
        self.library2.search.return_value.get.return_value = result2

        result = self.core.library.search({"any": ["a"]})

        assert result1 in result
        assert result2 in result
        self.library1.search.assert_called_once_with(
            query={"any": ["a"]}, uris=None, exact=False
        )
        self.library2.search.assert_called_once_with(
            query={"any": ["a"]}, uris=None, exact=False
        )

    def test_search_normalises_bad_queries(self):
        self.core.library.search({"any": "foobar"})
        self.library1.search.assert_called_once_with(
            query={"any": ["foobar"]}, uris=None, exact=False
        )


class LegacyFindExactToSearchLibraryTest(unittest.TestCase):
    def setUp(self):  # noqa: N802
        self.backend = mock.Mock()
        self.backend.actor_ref.actor_class.__name__ = "DummyBackend"
        self.backend.uri_schemes.get.return_value = ["dummy"]
        self.backend.library = mock.Mock(spec=backend.LibraryProvider)
        self.core = core.Core(mixer=None, backends=[self.backend])

    def test_core_search_call_backend_search_with_exact(self):
        self.core.library.search(query={"any": ["a"]})
        self.backend.library.search.assert_called_once_with(
            query=dict(any=["a"]), uris=None, exact=False
        )

    def test_core_search_with_exact_call_backend_search_with_exact(self):
        self.core.library.search(query={"any": ["a"]}, exact=True)
        self.backend.library.search.assert_called_once_with(
            query=dict(any=["a"]), uris=None, exact=True
        )

    def test_core_search_with_handles_legacy_backend(self):
        self.backend.library.search.return_value.get.side_effect = TypeError
        self.core.library.search(query={"any": ["a"]}, exact=True)
        # We are just testing that this doesn't fail.


class MockBackendCoreLibraryBase(unittest.TestCase):
    def setUp(self):  # noqa: N802
        dummy_root = Ref.directory(uri="dummy:directory", name="dummy")

        self.library = mock.Mock(spec=backend.LibraryProvider)
        self.library.root_directory.get.return_value = dummy_root

        self.backend = mock.Mock()
        self.backend.actor_ref.actor_class.__name__ = "DummyBackend"
        self.backend.uri_schemes.get.return_value = ["dummy"]
        self.backend.library = self.library

        self.core = core.Core(mixer=None, backends=[self.backend])


@mock.patch("mopidy.core.library.logger")
class BrowseBadBackendTest(MockBackendCoreLibraryBase):
    def test_backend_raises_exception_for_root(self, logger):
        # Might happen if root_directory is a property for some weird reason.
        self.library.root_directory.get.side_effect = Exception
        assert [] == self.core.library.browse(None)
        logger.exception.assert_called_with(mock.ANY, "DummyBackend")

    def test_backend_returns_none_for_root(self, logger):
        self.library.root_directory.get.return_value = None
        assert [] == self.core.library.browse(None)
        logger.error.assert_called_with(mock.ANY, "DummyBackend", mock.ANY)

    def test_backend_returns_wrong_type_for_root(self, logger):
        self.library.root_directory.get.return_value = 123
        assert [] == self.core.library.browse(None)
        logger.error.assert_called_with(mock.ANY, "DummyBackend", mock.ANY)

    def test_backend_raises_exception_for_browse(self, logger):
        self.library.browse.return_value.get.side_effect = Exception
        assert [] == self.core.library.browse("dummy:directory")
        logger.exception.assert_called_with(mock.ANY, "DummyBackend")

    def test_backend_returns_wrong_type_for_browse(self, logger):
        self.library.browse.return_value.get.return_value = [123]
        assert [] == self.core.library.browse("dummy:directory")
        logger.error.assert_called_with(mock.ANY, "DummyBackend", mock.ANY)


@mock.patch("mopidy.core.library.logger")
class GetDistinctBadBackendTest(MockBackendCoreLibraryBase):
    def test_backend_raises_exception(self, logger):
        self.library.get_distinct.return_value.get.side_effect = Exception
        assert set() == self.core.library.get_distinct("artist")
        logger.exception.assert_called_with(mock.ANY, "DummyBackend")

    def test_backend_returns_none(self, logger):
        self.library.get_distinct.return_value.get.return_value = None
        assert set() == self.core.library.get_distinct("artist")
        assert not logger.error.called

    def test_backend_returns_wrong_type(self, logger):
        self.library.get_distinct.return_value.get.return_value = "abc"
        assert set() == self.core.library.get_distinct("artist")
        logger.error.assert_called_with(mock.ANY, "DummyBackend", mock.ANY)

    def test_backend_returns_iterable_containing_wrong_types(self, logger):
        self.library.get_distinct.return_value.get.return_value = [1, 2, 3]
        assert set() == self.core.library.get_distinct("artist")
        logger.error.assert_called_with(mock.ANY, "DummyBackend", mock.ANY)


@mock.patch("mopidy.core.library.logger")
class GetImagesBadBackendTest(MockBackendCoreLibraryBase):
    def test_backend_raises_exception(self, logger):
        uri = "dummy:/1"
        self.library.get_images.return_value.get.side_effect = Exception
        assert {uri: ()} == self.core.library.get_images([uri])
        logger.exception.assert_called_with(mock.ANY, "DummyBackend")

    def test_backend_returns_none(self, logger):
        uri = "dummy:/1"
        self.library.get_images.return_value.get.return_value = None
        assert {uri: ()} == self.core.library.get_images([uri])
        assert not logger.error.called

    def test_backend_returns_wrong_type(self, logger):
        uri = "dummy:/1"
        self.library.get_images.return_value.get.return_value = "abc"
        assert {uri: ()} == self.core.library.get_images([uri])
        logger.error.assert_called_with(mock.ANY, "DummyBackend", mock.ANY)

    def test_backend_returns_mapping_containing_wrong_types(self, logger):
        uri = "dummy:/1"
        self.library.get_images.return_value.get.return_value = {uri: "abc"}
        assert {uri: ()} == self.core.library.get_images([uri])
        logger.error.assert_called_with(mock.ANY, "DummyBackend", mock.ANY)

    def test_backend_returns_mapping_containing_none(self, logger):
        uri = "dummy:/1"
        self.library.get_images.return_value.get.return_value = {uri: None}
        assert {uri: ()} == self.core.library.get_images([uri])
        logger.error.assert_called_with(mock.ANY, "DummyBackend", mock.ANY)

    def test_backend_returns_unknown_uri(self, logger):
        uri = "dummy:/1"
        self.library.get_images.return_value.get.return_value = {"foo": []}
        assert {uri: ()} == self.core.library.get_images([uri])
        logger.error.assert_called_with(mock.ANY, "DummyBackend", mock.ANY)


@mock.patch("mopidy.core.library.logger")
class LookupByUrisBadBackendTest(MockBackendCoreLibraryBase):
    def test_backend_raises_exception(self, logger):
        uri = "dummy:/1"
        self.library.lookup.return_value.get.side_effect = Exception
        assert {uri: []} == self.core.library.lookup(uris=[uri])
        logger.exception.assert_called_with(mock.ANY, "DummyBackend")

    def test_backend_returns_none(self, logger):
        uri = "dummy:/1"
        self.library.lookup.return_value.get.return_value = None
        assert {uri: []} == self.core.library.lookup(uris=[uri])
        assert not logger.error.called

    def test_backend_returns_wrong_type(self, logger):
        uri = "dummy:/1"
        self.library.lookup.return_value.get.return_value = "abc"
        assert {uri: []} == self.core.library.lookup(uris=[uri])
        logger.error.assert_called_with(mock.ANY, "DummyBackend", mock.ANY)

    def test_backend_returns_iterable_containing_wrong_types(self, logger):
        uri = "dummy:/1"
        self.library.lookup.return_value.get.return_value = [123]
        assert {uri: []} == self.core.library.lookup(uris=[uri])
        logger.error.assert_called_with(mock.ANY, "DummyBackend", mock.ANY)


@mock.patch("mopidy.core.library.logger")
class RefreshBadBackendTest(MockBackendCoreLibraryBase):
    def test_backend_raises_exception(self, logger):
        self.library.refresh.return_value.get.side_effect = Exception
        self.core.library.refresh()
        logger.exception.assert_called_with(mock.ANY, "DummyBackend")

    def test_backend_raises_exception_with_uri(self, logger):
        self.library.refresh.return_value.get.side_effect = Exception
        self.core.library.refresh("dummy:/1")
        logger.exception.assert_called_with(mock.ANY, "DummyBackend")


@mock.patch("mopidy.core.library.logger")
class SearchBadBackendTest(MockBackendCoreLibraryBase):
    def test_backend_raises_exception(self, logger):
        self.library.search.return_value.get.side_effect = Exception
        assert [] == self.core.library.search(query={"any": ["foo"]})
        logger.exception.assert_called_with(mock.ANY, "DummyBackend")

    def test_backend_raises_lookuperror(self, logger):
        # TODO: is this behavior desired? Do we need to continue handling
        # LookupError case specially.
        self.library.search.return_value.get.side_effect = LookupError
        with self.assertRaises(LookupError):
            self.core.library.search(query={"any": ["foo"]})

    def test_backend_returns_none(self, logger):
        self.library.search.return_value.get.return_value = None
        assert [] == self.core.library.search(query={"any": ["foo"]})
        assert not logger.error.called

    def test_backend_returns_wrong_type(self, logger):
        self.library.search.return_value.get.return_value = "abc"
        assert [] == self.core.library.search(query={"any": ["foo"]})
        logger.error.assert_called_with(mock.ANY, "DummyBackend", mock.ANY)



import unittest
from unittest import mock

from mopidy import backend, core
from mopidy.internal.models import TracklistState
from mopidy.models import TlTrack, Track


class TracklistTest(unittest.TestCase):
    def setUp(self):  # noqa:
        config = {"core": {"max_tracklist_length": 10000}}

        self.tracks = [
            Track(uri="dummy1:a", name="foo"),
            Track(uri="dummy1:b", name="foo"),
            Track(uri="dummy1:c", name="bar"),
        ]

        def lookup(uri):
            future = mock.Mock()
            future.get.return_value = [t for t in self.tracks if t.uri == uri]
            return future

        self.backend = mock.Mock()
        self.backend.uri_schemes.get.return_value = ["dummy1"]
        self.library = mock.Mock(spec=backend.LibraryProvider)
        self.library.lookup.side_effect = lookup
        self.backend.library = self.library

        self.core = core.Core(config, mixer=None, backends=[self.backend])
        self.tl_tracks = self.core.tracklist.add(
            uris=[t.uri for t in self.tracks]
        )

    def test_add_by_uri_looks_up_uri_in_library(self):
        self.library.lookup.reset_mock()
        self.core.tracklist.clear()

        tl_tracks = self.core.tracklist.add(uris=["dummy1:a"])

        self.library.lookup.assert_called_once_with("dummy1:a")
        assert 1 == len(tl_tracks)
        assert self.tracks[0] == tl_tracks[0].track
        assert tl_tracks == self.core.tracklist.get_tl_tracks()[(-1):]

    def test_add_by_uris_looks_up_uris_in_library(self):
        self.library.lookup.reset_mock()
        self.core.tracklist.clear()

        tl_tracks = self.core.tracklist.add(uris=[t.uri for t in self.tracks])

        self.library.lookup.assert_has_calls(
            [
                mock.call("dummy1:a"),
                mock.call("dummy1:b"),
                mock.call("dummy1:c"),
            ]
        )
        assert 3 == len(tl_tracks)
        assert self.tracks[0] == tl_tracks[0].track
        assert self.tracks[1] == tl_tracks[1].track
        assert self.tracks[2] == tl_tracks[2].track
        assert (
            tl_tracks
            == self.core.tracklist.get_tl_tracks()[(-len(tl_tracks)) :]
        )

    def test_remove_removes_tl_tracks_matching_query(self):
        tl_tracks = self.core.tracklist.remove({"name": ["foo"]})

        assert 2 == len(tl_tracks)
        self.assertListEqual(self.tl_tracks[:2], tl_tracks)

        assert 1 == self.core.tracklist.get_length()
        self.assertListEqual(
            self.tl_tracks[2:], self.core.tracklist.get_tl_tracks()
        )

    def test_remove_works_with_dict_instead_of_kwargs(self):
        tl_tracks = self.core.tracklist.remove({"name": ["foo"]})

        assert 2 == len(tl_tracks)
        self.assertListEqual(self.tl_tracks[:2], tl_tracks)

        assert 1 == self.core.tracklist.get_length()
        self.assertListEqual(
            self.tl_tracks[2:], self.core.tracklist.get_tl_tracks()
        )

    def test_filter_returns_tl_tracks_matching_query(self):
        tl_tracks = self.core.tracklist.filter({"name": ["foo"]})

        assert 2 == len(tl_tracks)
        self.assertListEqual(self.tl_tracks[:2], tl_tracks)

    def test_filter_works_with_dict_instead_of_kwargs(self):
        tl_tracks = self.core.tracklist.filter({"name": ["foo"]})

        assert 2 == len(tl_tracks)
        self.assertListEqual(self.tl_tracks[:2], tl_tracks)

    def test_filter_fails_if_values_isnt_iterable(self):
        with self.assertRaises(ValueError):
            self.core.tracklist.filter({"tlid": 3})

    def test_filter_fails_if_values_is_a_string(self):
        with self.assertRaises(ValueError):
            self.core.tracklist.filter({"uri": "a"})

    # TODO Extract tracklist tests from the local backend tests


class TracklistIndexTest(unittest.TestCase):
    def setUp(self):  # noqa: N802
        config = {"core": {"max_tracklist_length": 10000}}

        self.tracks = [
            Track(uri="dummy1:a", name="foo"),
            Track(uri="dummy1:b", name="foo"),
            Track(uri="dummy1:c", name="bar"),
        ]

        def lookup(uris):
            return {u: [t for t in self.tracks if t.uri == u] for u in uris}

        self.core = core.Core(config, mixer=None, backends=[])
        self.core.library = mock.Mock(spec=core.LibraryController)
        self.core.library.lookup.side_effect = lookup

        self.core.playback = mock.Mock(spec=core.PlaybackController)

        self.tl_tracks = self.core.tracklist.add(
            uris=[t.uri for t in self.tracks]
        )

    def test_index_returns_index_of_track(self):
        assert 0 == self.core.tracklist.index(self.tl_tracks[0])
        assert 1 == self.core.tracklist.index(self.tl_tracks[1])
        assert 2 == self.core.tracklist.index(self.tl_tracks[2])

    def test_index_returns_none_if_item_not_found(self):
        tl_track = TlTrack(0, Track())
        assert self.core.tracklist.index(tl_track) is None

    def test_index_returns_none_if_called_with_none(self):
        assert self.core.tracklist.index(None) is None

    def test_index_errors_out_for_invalid_tltrack(self):
        with self.assertRaises(ValueError):
            self.core.tracklist.index("abc")

    def test_index_return_index_when_called_with_tlids(self):
        tl_tracks = self.tl_tracks
        assert 0 == self.core.tracklist.index(tlid=tl_tracks[0].tlid)
        assert 1 == self.core.tracklist.index(tlid=tl_tracks[1].tlid)
        assert 2 == self.core.tracklist.index(tlid=tl_tracks[2].tlid)

    def test_index_returns_none_if_tlid_not_found(self):
        assert self.core.tracklist.index(tlid=123) is None

    def test_index_returns_none_if_called_with_tlid_none(self):
        assert self.core.tracklist.index(tlid=None) is None

    def test_index_errors_out_for_invalid_tlid(self):
        with self.assertRaises(ValueError):
            self.core.tracklist.index(tlid=-1)

    def test_index_without_args_returns_current_tl_track_index(self):
        self.core.playback.get_current_tl_track.side_effect = [
            None,
            self.tl_tracks[0],
            self.tl_tracks[1],
            self.tl_tracks[2],
        ]

        assert self.core.tracklist.index() is None
        assert 0 == self.core.tracklist.index()
        assert 1 == self.core.tracklist.index()
        assert 2 == self.core.tracklist.index()


class TracklistSaveLoadStateTest(unittest.TestCase):
    def setUp(self):  # noqa: N802
        config = {"core": {"max_tracklist_length": 10000}}

        self.tracks = [
            Track(uri="dummy1:a", name="foo"),
            Track(uri="dummy1:b", name="foo"),
            Track(uri="dummy1:c", name="bar"),
        ]

        self.tl_tracks = [
            TlTrack(tlid=4, track=Track(uri="first", name="First")),
            TlTrack(tlid=5, track=Track(uri="second", name="Second")),
            TlTrack(tlid=6, track=Track(uri="third", name="Third")),
            TlTrack(tlid=8, track=Track(uri="last", name="Last")),
        ]

        def lookup(uris):
            return {u: [t for t in self.tracks if t.uri == u] for u in uris}

        self.core = core.Core(config, mixer=None, backends=[])
        self.core.library = mock.Mock(spec=core.LibraryController)
        self.core.library.lookup.side_effect = lookup

        self.core.playback = mock.Mock(spec=core.PlaybackController)

    def test_save(self):
        tl_tracks = self.core.tracklist.add(uris=[t.uri for t in self.tracks])
        consume = True
        next_tlid = len(tl_tracks) + 1
        self.core.tracklist.set_consume(consume)
        target = TracklistState(
            consume=consume,
            repeat=False,
            single=False,
            random=False,
            next_tlid=next_tlid,
            tl_tracks=tl_tracks,
        )
        value = self.core.tracklist._save_state()
        assert target == value

    def test_load(self):
        old_version = self.core.tracklist.get_version()
        target = TracklistState(
            consume=False,
            repeat=True,
            single=True,
            random=False,
            next_tlid=12,
            tl_tracks=self.tl_tracks,
        )
        coverage = ["mode", "tracklist"]
        self.core.tracklist._load_state(target, coverage)
        assert self.core.tracklist.get_consume() is False
        assert self.core.tracklist.get_repeat() is True
        assert self.core.tracklist.get_single() is True
        assert self.core.tracklist.get_random() is False
        assert 12 == self.core.tracklist._next_tlid
        assert 4 == self.core.tracklist.get_length()
        assert self.tl_tracks == self.core.tracklist.get_tl_tracks()
        assert self.core.tracklist.get_version() > old_version

        # after load, adding more tracks must be possible
        self.core.tracklist.add(uris=[self.tracks[1].uri])
        assert 13 == self.core.tracklist._next_tlid
        assert 5 == self.core.tracklist.get_length()

    def test_load_mode_only(self):
        old_version = self.core.tracklist.get_version()
        target = TracklistState(
            consume=False,
            repeat=True,
            single=True,
            random=False,
            next_tlid=12,
            tl_tracks=self.tl_tracks,
        )
        coverage = ["mode"]
        self.core.tracklist._load_state(target, coverage)
        assert self.core.tracklist.get_consume() is False
        assert self.core.tracklist.get_repeat() is True
        assert self.core.tracklist.get_single() is True
        assert self.core.tracklist.get_random() is False
        assert 1 == self.core.tracklist._next_tlid
        assert 0 == self.core.tracklist.get_length()
        assert [] == self.core.tracklist.get_tl_tracks()
        assert self.core.tracklist.get_version() == old_version

    def test_load_tracklist_only(self):
        old_version = self.core.tracklist.get_version()
        target = TracklistState(
            consume=False,
            repeat=True,
            single=True,
            random=False,
            next_tlid=12,
            tl_tracks=self.tl_tracks,
        )
        coverage = ["tracklist"]
        self.core.tracklist._load_state(target, coverage)
        assert self.core.tracklist.get_consume() is False
        assert self.core.tracklist.get_repeat() is False
        assert self.core.tracklist.get_single() is False
        assert self.core.tracklist.get_random() is False
        assert 12 == self.core.tracklist._next_tlid
        assert 4 == self.core.tracklist.get_length()
        assert self.tl_tracks == self.core.tracklist.get_tl_tracks()
        assert self.core.tracklist.get_version() > old_version

    def test_load_invalid_type(self):
        with self.assertRaises(TypeError):
            self.core.tracklist._load_state(11, None)

    def test_load_none(self):
        self.core.tracklist._load_state(None, None)



import os
import pathlib
import urllib

from mopidy import models
from mopidy.internal import path

from . import Extension


def path_to_uri(path, scheme=Extension.ext_name):
    """Convert file path to URI."""
    bytes_path = os.path.normpath(bytes(path))
    uripath = urllib.parse.quote_from_bytes(bytes_path)
    return urllib.parse.urlunsplit((scheme, None, uripath, None, None))


def uri_to_path(uri):
    """Convert URI to file path."""
    return path.uri_to_path(uri)


def name_from_path(path):
    """Extract name from file path."""
    name = bytes(pathlib.Path(path.stem))
    try:
        return name.decode(errors="replace")
    except UnicodeError:
        return None


def path_from_name(name, ext=None, sep="|"):
    """Convert name with optional extension to file path."""
    if ext:
        name = name.replace(os.sep, sep) + ext
    else:
        name = name.replace(os.sep, sep)
    return pathlib.Path(name)


def path_to_ref(path):
    return models.Ref.playlist(uri=path_to_uri(path), name=name_from_path(path))


def load_items(fp, basedir):
    refs = []
    name = None
    for line in filter(None, (line.strip() for line in fp)):
        if line.startswith("#"):
            if line.startswith("#EXTINF:"):
                name = line.partition(",")[2]
            continue
        elif not urllib.parse.urlsplit(line).scheme:
            path = basedir / line
            if not name:
                name = name_from_path(path)
            uri = path_to_uri(path, scheme="file")
        else:
            # TODO: ensure this is urlencoded
            uri = line  # do *not* extract name from (stream?) URI path
        refs.append(models.Ref.track(uri=uri, name=name))
        name = None
    return refs


def dump_items(items, fp):
    if any(item.name for item in items):
        print("#EXTM3U", file=fp)
    for item in items:
        if item.name:
            print(f"#EXTINF:-1,{item.name}", file=fp)
        # TODO: convert file URIs to (relative) paths?
        if isinstance(item.uri, bytes):
            print(item.uri.decode(), file=fp)
        else:
            print(item.uri, file=fp)


def playlist(path, items=None, mtime=None):
    if items is None:
        items = []
    return models.Playlist(
        uri=path_to_uri(path),
        name=name_from_path(path),
        tracks=[models.Track(uri=item.uri, name=item.name) for item in items],
        last_modified=(int(mtime * 1000) if mtime else None),
    )



import logging
import os

import mopidy
from mopidy import config as config_lib
from mopidy import exceptions, ext

logger = logging.getLogger(__name__)


class Extension(ext.Extension):
    dist_name = "Mopidy-HTTP"
    ext_name = "http"
    version = mopidy.__version__

    def get_default_config(self):
        conf_file = os.path.join(os.path.dirname(__file__), "ext.conf")
        return config_lib.read(conf_file)

    def get_config_schema(self):
        schema = super().get_config_schema()
        schema["hostname"] = config_lib.Hostname()
        schema["port"] = config_lib.Port()
        schema["static_dir"] = config_lib.Deprecated()
        schema["zeroconf"] = config_lib.String(optional=True)
        schema["allowed_origins"] = config_lib.List(optional=True)
        schema["csrf_protection"] = config_lib.Boolean(optional=True)
        schema["default_app"] = config_lib.String(optional=True)
        return schema

    def validate_environment(self):
        try:
            import tornado.web  # noqa
        except ImportError as e:
            raise exceptions.ExtensionError("tornado library not found", e)

    def setup(self, registry):
        from .actor import HttpFrontend
        from .handlers import make_mopidy_app_factory

        HttpFrontend.apps = registry["http:app"]
        HttpFrontend.statics = registry["http:static"]

        registry.add("frontend", HttpFrontend)
        registry.add(
            "http:app",
            {
                "name": "mopidy",
                "factory": make_mopidy_app_factory(
                    registry["http:app"], registry["http:static"]
                ),
            },
        )



import platform
import sys
import warnings

import pkg_resources

if not sys.version_info >= (3, 7):
    sys.exit(
        f"ERROR: Mopidy requires Python >= 3.7, "
        f"but found {platform.python_version()}."
    )

warnings.filterwarnings("ignore", "could not open display")

__version__ = pkg_resources.get_distribution("Mopidy").version



import unittest

from mopidy.models import ImmutableObject


class Model(ImmutableObject):
    uri = None
    name = None
    models = frozenset()

    def __init__(self, *args, **kwargs):
        self.__dict__["models"] = frozenset(kwargs.pop("models", None) or [])
        super().__init__(self, *args, **kwargs)


class SubModel(ImmutableObject):
    uri = None
    name = None


class GenericCopyTest(unittest.TestCase):
    def compare(self, orig, other):
        assert orig == other
        assert id(orig) != id(other)

    def test_copying_model(self):
        model = Model()
        self.compare(model, model.replace())

    def test_copying_model_with_basic_values(self):
        model = Model(name="foo", uri="bar")
        other = model.replace(name="baz")
        assert "baz" == other.name
        assert "bar" == other.uri

    def test_copying_model_with_missing_values(self):
        model = Model(uri="bar")
        other = model.replace(name="baz")
        assert "baz" == other.name
        assert "bar" == other.uri

    def test_copying_model_with_private_internal_value(self):
        model = Model(models=[SubModel(name=123)])
        other = model.replace(models=[SubModel(name=345)])
        assert SubModel(name=345) in other.models

    def test_copying_model_with_invalid_key(self):
        with self.assertRaises(TypeError):
            Model().replace(invalid_key=True)

    def test_copying_model_to_remove(self):
        model = Model(name="foo").replace(name=None)
        assert model == Model()


class ModelTest(unittest.TestCase):
    def test_uri(self):
        uri = "an_uri"
        model = Model(uri=uri)
        assert model.uri == uri
        with self.assertRaises(AttributeError):
            model.uri = None

    def test_name(self):
        name = "a name"
        model = Model(name=name)
        assert model.name == name
        with self.assertRaises(AttributeError):
            model.name = None

    def test_submodels(self):
        models = [SubModel(name=123), SubModel(name=456)]
        model = Model(models=models)
        assert set(model.models) == set(models)
        with self.assertRaises(AttributeError):
            model.models = None

    def test_models_none(self):
        assert set() == Model(models=None).models

    def test_invalid_kwarg(self):
        with self.assertRaises(TypeError):
            Model(foo="baz")

    def test_repr_without_models(self):
        assert "Model(name='name', uri='uri')" == repr(
            Model(uri="uri", name="name")
        )

    def test_repr_with_models(self):
        assert (
            "Model(models=[SubModel(name=123)], name='name', uri='uri')"
            == repr(Model(uri="uri", name="name", models=[SubModel(name=123)]))
        )

    def test_serialize_without_models(self):
        self.assertDictEqual(
            {"__model__": "Model", "uri": "uri", "name": "name"},
            Model(uri="uri", name="name").serialize(),
        )

    def test_serialize_with_models(self):
        submodel = SubModel(name=123)
        self.assertDictEqual(
            {
                "__model__": "Model",
                "uri": "uri",
                "name": "name",
                "models": [submodel.serialize()],
            },
            Model(uri="uri", name="name", models=[submodel]).serialize(),
        )

    def test_eq_uri(self):
        model1 = Model(uri="uri1")
        model2 = Model(uri="uri1")
        assert model1 == model2
        assert hash(model1) == hash(model2)

    def test_eq_name(self):
        model1 = Model(name="name1")
        model2 = Model(name="name1")
        assert model1 == model2
        assert hash(model1) == hash(model2)

    def test_eq_models(self):
        models = [SubModel()]
        model1 = Model(models=models)
        model2 = Model(models=models)
        assert model1 == model2
        assert hash(model1) == hash(model2)

    def test_eq_models_order(self):
        submodel1 = SubModel(name="name1")
        submodel2 = SubModel(name="name2")
        model1 = Model(models=[submodel1, submodel2])
        model2 = Model(models=[submodel2, submodel1])
        assert model1 == model2
        assert hash(model1) == hash(model2)

    def test_eq_none(self):
        assert Model() is not None

    def test_eq_other(self):
        assert Model() != "other"

    def test_ne_uri(self):
        model1 = Model(uri="uri1")
        model2 = Model(uri="uri2")
        assert model1 != model2
        assert hash(model1) != hash(model2)

    def test_ne_name(self):
        model1 = Model(name="name1")
        model2 = Model(name="name2")
        assert model1 != model2
        assert hash(model1) != hash(model2)

    def test_ne_models(self):
        model1 = Model(models=[SubModel(name="name1")])
        model2 = Model(models=[SubModel(name="name2")])
        assert model1 != model2
        assert hash(model1) != hash(model2)

    def test_ignores_values_with_default_value_none(self):
        model1 = Model(name="name1")
        model2 = Model(name="name1", uri=None)
        assert model1 == model2
        assert hash(model1) == hash(model2)



import json
import unittest
from unittest import mock

import pykka

from mopidy import core, models
from mopidy.internal import deprecation, jsonrpc

from tests import dummy_backend


class Calculator:
    def __init__(self):
        self._mem = None

    def model(self):
        return "TI83"

    def add(self, a, b):
        """Returns the sum of the given numbers"""
        return a + b

    def sub(self, a, b):
        return a - b

    def set_mem(self, value):
        self._mem = value

    def get_mem(self):
        return self._mem

    def describe(self):
        return {
            "add": "Returns the sum of the terms",
            "sub": "Returns the diff of the terms",
        }

    def take_it_all(self, a, b, c=True, *args, **kwargs):
        pass

    def _secret(self):
        return "Grand Unified Theory"

    def fail(self):
        raise ValueError("What did you expect?")


class JsonRpcTestBase(unittest.TestCase):
    def setUp(self):  # noqa: N802
        self.backend = dummy_backend.create_proxy()
        self.calc = Calculator()

        with deprecation.ignore():
            self.core = core.Core.start(backends=[self.backend]).proxy()

        self.jrw = jsonrpc.JsonRpcWrapper(
            objects={
                "hello": lambda: "Hello, world!",
                "calc": self.calc,
                "core": self.core,
                "core.playback": self.core.playback,
                "core.tracklist": self.core.tracklist,
                "get_uri_schemes": self.core.get_uri_schemes,
            },
            encoders=[models.ModelJSONEncoder],
            decoders=[models.model_json_decoder],
        )

    def tearDown(self):  # noqa: N802
        pykka.ActorRegistry.stop_all()


class JsonRpcSetupTest(JsonRpcTestBase):
    def test_empty_object_mounts_is_not_allowed(self):
        with self.assertRaises(AttributeError):
            jsonrpc.JsonRpcWrapper(objects={"": Calculator()})


class JsonRpcSerializationTest(JsonRpcTestBase):
    def test_handle_json_converts_from_and_to_json(self):
        self.jrw.handle_data = mock.Mock()
        self.jrw.handle_data.return_value = {"foo": "response"}

        request = '{"foo": "request"}'
        response = self.jrw.handle_json(request)

        self.jrw.handle_data.assert_called_once_with({"foo": "request"})
        assert response == '{"foo": "response"}'

    def test_handle_json_decodes_mopidy_models(self):
        self.jrw.handle_data = mock.Mock()
        self.jrw.handle_data.return_value = []

        request = '{"foo": {"__model__": "Artist", "name": "bar"}}'
        self.jrw.handle_json(request)

        self.jrw.handle_data.assert_called_once_with(
            {"foo": models.Artist(name="bar")}
        )

    def test_handle_json_encodes_mopidy_models(self):
        self.jrw.handle_data = mock.Mock()
        self.jrw.handle_data.return_value = {"foo": models.Artist(name="bar")}

        request = "[]"
        response = json.loads(self.jrw.handle_json(request))

        assert "foo" in response
        assert "__model__" in response["foo"]
        assert response["foo"]["__model__"] == "Artist"
        assert "name" in response["foo"]
        assert response["foo"]["name"] == "bar"

    def test_handle_json_returns_nothing_for_notices(self):
        request = '{"jsonrpc": "2.0", "method": "core.get_uri_schemes"}'
        response = self.jrw.handle_json(request)

        assert response is None

    def test_invalid_json_command_causes_parse_error(self):
        request = '{"jsonrpc": "2.0", "method": "foobar, "params": "bar", "baz]'
        response = self.jrw.handle_json(request)
        response = json.loads(response)

        assert response["jsonrpc"] == "2.0"
        error = response["error"]
        assert error["code"] == (-32700)
        assert error["message"] == "Parse error"

    def test_invalid_json_batch_causes_parse_error(self):
        request = """[
            {"jsonrpc": "2.0", "method": "sum", "params": [1,2,4], "id": "1"},
            {"jsonrpc": "2.0", "method"
        ]"""
        response = self.jrw.handle_json(request)
        response = json.loads(response)

        assert response["jsonrpc"] == "2.0"
        error = response["error"]
        assert error["code"] == (-32700)
        assert error["message"] == "Parse error"


class JsonRpcSingleCommandTest(JsonRpcTestBase):
    def test_call_method_on_root(self):
        request = {
            "jsonrpc": "2.0",
            "method": "hello",
            "id": 1,
        }
        response = self.jrw.handle_data(request)

        assert response["jsonrpc"] == "2.0"
        assert response["id"] == 1
        assert "error" not in response
        assert response["result"] == "Hello, world!"

    def test_call_method_on_plain_object(self):
        request = {
            "jsonrpc": "2.0",
            "method": "calc.model",
            "id": 1,
        }
        response = self.jrw.handle_data(request)

        assert response["jsonrpc"] == "2.0"
        assert response["id"] == 1
        assert "error" not in response
        assert response["result"] == "TI83"

    def test_call_method_which_returns_dict_from_plain_object(self):
        request = {
            "jsonrpc": "2.0",
            "method": "calc.describe",
            "id": 1,
        }
        response = self.jrw.handle_data(request)

        assert response["jsonrpc"] == "2.0"
        assert response["id"] == 1
        assert "error" not in response
        assert "add" in response["result"]
        assert "sub" in response["result"]

    def test_call_method_on_actor_root(self):
        request = {
            "jsonrpc": "2.0",
            "method": "core.get_uri_schemes",
            "id": 1,
        }
        response = self.jrw.handle_data(request)

        assert response["jsonrpc"] == "2.0"
        assert response["id"] == 1
        assert "error" not in response
        assert response["result"] == ["dummy"]

    def test_call_method_on_actor_member(self):
        request = {
            "jsonrpc": "2.0",
            "method": "core.playback.get_time_position",
            "id": 1,
        }
        response = self.jrw.handle_data(request)

        assert response["result"] == 0

    def test_call_method_which_is_a_directly_mounted_actor_member(self):
        # 'get_uri_schemes' isn't a regular callable, but a Pykka
        # CallableProxy. This test checks that CallableProxy objects are
        # threated by JsonRpcWrapper like any other callable.

        request = {
            "jsonrpc": "2.0",
            "method": "get_uri_schemes",
            "id": 1,
        }
        response = self.jrw.handle_data(request)

        assert response["jsonrpc"] == "2.0"
        assert response["id"] == 1
        assert "error" not in response
        assert response["result"] == ["dummy"]

    def test_call_method_with_positional_params(self):
        request = {
            "jsonrpc": "2.0",
            "method": "calc.add",
            "params": [3, 4],
            "id": 1,
        }
        response = self.jrw.handle_data(request)

        assert response["result"] == 7

    def test_call_methods_with_named_params(self):
        request = {
            "jsonrpc": "2.0",
            "method": "calc.add",
            "params": {"a": 3, "b": 4},
            "id": 1,
        }
        response = self.jrw.handle_data(request)

        assert response["result"] == 7


class JsonRpcSingleNotificationTest(JsonRpcTestBase):
    def test_notification_does_not_return_a_result(self):
        request = {
            "jsonrpc": "2.0",
            "method": "core.get_uri_schemes",
        }
        response = self.jrw.handle_data(request)

        assert response is None

    def test_notification_makes_an_observable_change(self):
        assert self.calc.get_mem() is None

        request = {
            "jsonrpc": "2.0",
            "method": "calc.set_mem",
            "params": [37],
        }
        response = self.jrw.handle_data(request)

        assert response is None
        assert self.calc.get_mem() == 37

    def test_notification_unknown_method_returns_nothing(self):
        request = {
            "jsonrpc": "2.0",
            "method": "bogus",
            "params": ["bogus"],
        }
        response = self.jrw.handle_data(request)

        assert response is None


class JsonRpcBatchTest(JsonRpcTestBase):
    def test_batch_of_only_commands_returns_all(self):
        self.core.tracklist.set_random(True).get()

        request = [
            {"jsonrpc": "2.0", "method": "core.tracklist.get_repeat", "id": 1},
            {"jsonrpc": "2.0", "method": "core.tracklist.get_random", "id": 2},
            {"jsonrpc": "2.0", "method": "core.tracklist.get_single", "id": 3},
        ]
        response = self.jrw.handle_data(request)

        assert len(response) == 3

        response = {row["id"]: row for row in response}
        assert response[1]["result"] is False
        assert response[2]["result"] is True
        assert response[3]["result"] is False

    def test_batch_of_commands_and_notifications_returns_some(self):
        self.core.tracklist.set_random(True).get()

        request = [
            {"jsonrpc": "2.0", "method": "core.tracklist.get_repeat"},
            {"jsonrpc": "2.0", "method": "core.tracklist.get_random", "id": 2},
            {"jsonrpc": "2.0", "method": "core.tracklist.get_single", "id": 3},
        ]
        response = self.jrw.handle_data(request)

        assert len(response) == 2

        response = {row["id"]: row for row in response}
        assert 1 not in response
        assert response[2]["result"] is True
        assert response[3]["result"] is False

    def test_batch_of_only_notifications_returns_nothing(self):
        self.core.tracklist.set_random(True).get()

        request = [
            {"jsonrpc": "2.0", "method": "core.tracklist.get_repeat"},
            {"jsonrpc": "2.0", "method": "core.tracklist.get_random"},
            {"jsonrpc": "2.0", "method": "core.tracklist.get_single"},
        ]
        response = self.jrw.handle_data(request)

        assert response is None


class JsonRpcSingleCommandErrorTest(JsonRpcTestBase):
    def test_application_error_response(self):
        request = {
            "jsonrpc": "2.0",
            "method": "calc.fail",
            "params": [],
            "id": 1,
        }
        response = self.jrw.handle_data(request)

        assert "result" not in response

        error = response["error"]
        assert error["code"] == 0
        assert error["message"] == "Application error"

        data = error["data"]
        assert data["type"] == "ValueError"
        assert "What did you expect?" in data["message"]
        assert "traceback" in data
        assert "Traceback (most recent call last):" in data["traceback"]

    def test_missing_jsonrpc_member_causes_invalid_request_error(self):
        request = {
            "method": "core.get_uri_schemes",
            "id": 1,
        }
        response = self.jrw.handle_data(request)

        assert response["id"] is None
        error = response["error"]
        assert error["code"] == (-32600)
        assert error["message"] == "Invalid Request"
        assert error["data"] == "'jsonrpc' member must be included"

    def test_wrong_jsonrpc_version_causes_invalid_request_error(self):
        request = {
            "jsonrpc": "3.0",
            "method": "core.get_uri_schemes",
            "id": 1,
        }
        response = self.jrw.handle_data(request)

        assert response["id"] is None
        error = response["error"]
        assert error["code"] == (-32600)
        assert error["message"] == "Invalid Request"
        assert error["data"] == "'jsonrpc' value must be '2.0'"

    def test_missing_method_member_causes_invalid_request_error(self):
        request = {
            "jsonrpc": "2.0",
            "id": 1,
        }
        response = self.jrw.handle_data(request)

        assert response["id"] is None
        error = response["error"]
        assert error["code"] == (-32600)
        assert error["message"] == "Invalid Request"
        assert error["data"] == "'method' member must be included"

    def test_invalid_method_value_causes_invalid_request_error(self):
        request = {
            "jsonrpc": "2.0",
            "method": 1,
            "id": 1,
        }
        response = self.jrw.handle_data(request)

        assert response["id"] is None
        error = response["error"]
        assert error["code"] == (-32600)
        assert error["message"] == "Invalid Request"
        assert error["data"] == "'method' must be a string"

    def test_invalid_params_value_causes_invalid_request_error(self):
        request = {
            "jsonrpc": "2.0",
            "method": "core.get_uri_schemes",
            "params": "foobar",
            "id": 1,
        }
        response = self.jrw.handle_data(request)

        assert response["id"] is None
        error = response["error"]
        assert error["code"] == (-32600)
        assert error["message"] == "Invalid Request"
        assert (
            error["data"] == "'params', if given, must be an array or an object"
        )

    def test_method_on_without_object_causes_unknown_method_error(self):
        request = {
            "jsonrpc": "2.0",
            "method": "bogus",
            "id": 1,
        }
        response = self.jrw.handle_data(request)

        error = response["error"]
        assert error["code"] == (-32601)
        assert error["message"] == "Method not found"
        assert (
            error["data"]
            == "Could not find object mount in method name 'bogus'"
        )

    def test_method_on_unknown_object_causes_unknown_method_error(self):
        request = {
            "jsonrpc": "2.0",
            "method": "bogus.bogus",
            "id": 1,
        }
        response = self.jrw.handle_data(request)

        error = response["error"]
        assert error["code"] == (-32601)
        assert error["message"] == "Method not found"
        assert error["data"] == "No object found at 'bogus'"

    def test_unknown_method_on_known_object_causes_unknown_method_error(self):
        request = {
            "jsonrpc": "2.0",
            "method": "core.bogus",
            "id": 1,
        }
        response = self.jrw.handle_data(request)

        error = response["error"]
        assert error["code"] == (-32601)
        assert error["message"] == "Method not found"
        assert error["data"] == "Object mounted at 'core' has no member 'bogus'"

    def test_private_method_causes_unknown_method_error(self):
        request = {
            "jsonrpc": "2.0",
            "method": "core._secret",
            "id": 1,
        }
        response = self.jrw.handle_data(request)

        error = response["error"]
        assert error["code"] == (-32601)
        assert error["message"] == "Method not found"
        assert error["data"] == "Private methods are not exported"

    def test_invalid_params_causes_invalid_params_error(self):
        request = {
            "jsonrpc": "2.0",
            "method": "core.get_uri_schemes",
            "params": ["bogus"],
            "id": 1,
        }
        response = self.jrw.handle_data(request)

        error = response["error"]
        assert error["code"] == (-32602)
        assert error["message"] == "Invalid params"

        data = error["data"]
        assert data["type"] == "TypeError"
        assert (
            "get_uri_schemes() takes 1 positional argument but 2 were given"
            in data["message"]
        )
        assert "traceback" in data
        assert "Traceback (most recent call last):" in data["traceback"]


class JsonRpcBatchErrorTest(JsonRpcTestBase):
    def test_empty_batch_list_causes_invalid_request_error(self):
        request = []
        response = self.jrw.handle_data(request)

        assert response["id"] is None
        error = response["error"]
        assert error["code"] == (-32600)
        assert error["message"] == "Invalid Request"
        assert error["data"] == "Batch list cannot be empty"

    def test_batch_with_invalid_command_causes_invalid_request_error(self):
        request = [1]
        response = self.jrw.handle_data(request)

        assert len(response) == 1
        response = response[0]
        assert response["id"] is None
        error = response["error"]
        assert error["code"] == (-32600)
        assert error["message"] == "Invalid Request"
        assert error["data"] == "Request must be an object"

    def test_batch_with_invalid_commands_causes_invalid_request_error(self):
        request = [1, 2, 3]
        response = self.jrw.handle_data(request)

        assert len(response) == 3
        response = response[2]
        assert response["id"] is None
        error = response["error"]
        assert error["code"] == (-32600)
        assert error["message"] == "Invalid Request"
        assert error["data"] == "Request must be an object"

    def test_batch_of_both_successfull_and_failing_requests(self):
        request = [
            # Call with positional params
            {
                "jsonrpc": "2.0",
                "method": "core.playback.seek",
                "params": [47],
                "id": "1",
            },
            # Notification
            {
                "jsonrpc": "2.0",
                "method": "core.tracklist.set_consume",
                "params": [True],
            },
            # Call with positional params
            {
                "jsonrpc": "2.0",
                "method": "core.tracklist.set_repeat",
                "params": [False],
                "id": "2",
            },
            # Invalid request
            {"foo": "boo"},
            # Unknown method
            {
                "jsonrpc": "2.0",
                "method": "foo.get",
                "params": {"name": "myself"},
                "id": "5",
            },
            # Call without params
            {
                "jsonrpc": "2.0",
                "method": "core.tracklist.get_random",
                "id": "9",
            },
        ]
        response = self.jrw.handle_data(request)

        assert len(response) == 5
        response = {row["id"]: row for row in response}
        assert response["1"]["result"] is False
        assert response["2"]["result"] is None
        assert response[None]["error"]["code"] == (-32600)
        assert response["5"]["error"]["code"] == (-32601)
        assert response["9"]["result"] is False


class JsonRpcInspectorTest(JsonRpcTestBase):
    def test_empty_object_mounts_is_not_allowed(self):
        with self.assertRaises(AttributeError):
            jsonrpc.JsonRpcInspector(objects={"": Calculator})

    def test_can_describe_method_on_root(self):
        inspector = jsonrpc.JsonRpcInspector({"hello": lambda: "Hello, world!"})

        methods = inspector.describe()

        assert "hello" in methods
        assert len(methods["hello"]["params"]) == 0

    def test_inspector_can_describe_an_object_with_methods(self):
        inspector = jsonrpc.JsonRpcInspector({"calc": Calculator})

        methods = inspector.describe()

        assert "calc.add" in methods
        assert (
            methods["calc.add"]["description"]
            == "Returns the sum of the given numbers"
        )

        assert "calc.sub" in methods
        assert "calc.take_it_all" in methods
        assert "calc._secret" not in methods
        assert "calc.__init__" not in methods

        method = methods["calc.take_it_all"]
        assert "params" in method

        params = method["params"]

        assert params[0]["name"] == "a"
        assert "default" not in params[0]

        assert params[1]["name"] == "b"
        assert "default" not in params[1]

        assert params[2]["name"] == "c"
        assert params[2]["default"] is True

        assert params[3]["name"] == "args"
        assert "default" not in params[3]
        assert params[3]["varargs"] is True

        assert params[4]["name"] == "kwargs"
        assert "default" not in params[4]
        assert params[4]["kwargs"] is True

    def test_inspector_can_describe_a_bunch_of_large_classes(self):
        inspector = jsonrpc.JsonRpcInspector(
            {
                "core.get_uri_schemes": core.Core.get_uri_schemes,
                "core.library": core.LibraryController,
                "core.playback": core.PlaybackController,
                "core.playlists": core.PlaylistsController,
                "core.tracklist": core.TracklistController,
            }
        )

        methods = inspector.describe()

        assert "core.get_uri_schemes" in methods
        assert len(methods["core.get_uri_schemes"]["params"]) == 0

        assert "core.library.lookup" in methods
        assert methods["core.library.lookup"]["params"][0]["name"] == "uris"

        assert "core.playback.next" in methods
        assert len(methods["core.playback.next"]["params"]) == 0

        assert "core.playlists.as_list" in methods
        assert len(methods["core.playlists.as_list"]["params"]) == 0

        assert "core.tracklist.filter" in methods
        assert (
            methods["core.tracklist.filter"]["params"][0]["name"] == "criteria"
        )



import os

import mopidy
from mopidy import config, ext


class Extension(ext.Extension):

    dist_name = "Mopidy-Stream"
    ext_name = "stream"
    version = mopidy.__version__

    def get_default_config(self):
        conf_file = os.path.join(os.path.dirname(__file__), "ext.conf")
        return config.read(conf_file)

    def get_config_schema(self):
        schema = super().get_config_schema()
        schema["protocols"] = config.List()
        schema["metadata_blacklist"] = config.List(optional=True)
        schema["timeout"] = config.Integer(minimum=1000, maximum=1000 * 60 * 60)
        return schema

    def validate_environment(self):
        pass

    def setup(self, registry):
        from .actor import StreamBackend

        registry.add("backend", StreamBackend)



import collections
import datetime
import logging
import numbers

from mopidy.internal import log
from mopidy.internal.gi import GLib, Gst
from mopidy.models import Album, Artist, Track

logger = logging.getLogger(__name__)


def repr_tags(taglist, max_bytes=10):
    """Returns a printable representation of a :class:`Gst.TagList`.

    Tag values of type bytes are truncated to the specified length to avoid
    large amounts of output when logging.

    :param taglist: A GStreamer taglist to be represented.
    :type taglist: :class:`Gst.TagList`
    :param max_bytes: The maximum number of bytes to show for bytes tag values.
    :type max_bytes: int
    :rtype: string
    """
    result = dict(taglist)
    for tag_values in result.values():
        for i, val in enumerate(tag_values):
            if type(val) is bytes and len(val) > max_bytes:
                tag_values[i] = val[:max_bytes] + b"..."
    return repr(result)


def convert_taglist(taglist):
    """Convert a :class:`Gst.TagList` to plain Python types.

    Knows how to convert:

    - Dates
    - Buffers
    - Numbers
    - Strings
    - Booleans

    Unknown types will be ignored and trace logged. Tag keys are all strings
    defined as part GStreamer under GstTagList_.

    .. _GstTagList: https://developer.gnome.org/gstreamer/stable/\
gstreamer-GstTagList.html

    :param taglist: A GStreamer taglist to be converted.
    :type taglist: :class:`Gst.TagList`
    :rtype: dictionary of tag keys with a list of values.
    """
    result = collections.defaultdict(list)

    for n in range(taglist.n_tags()):
        tag = taglist.nth_tag_name(n)

        for i in range(taglist.get_tag_size(tag)):
            value = taglist.get_value_index(tag, i)

            if isinstance(value, GLib.Date):
                try:
                    date = datetime.date(
                        value.get_year(), value.get_month(), value.get_day()
                    )
                    result[tag].append(date.isoformat())
                except ValueError:
                    logger.debug(
                        "Ignoring dodgy date value: %d-%d-%d",
                        value.get_year(),
                        value.get_month(),
                        value.get_day(),
                    )
            elif isinstance(value, Gst.DateTime):
                result[tag].append(value.to_iso8601_string())
            elif isinstance(value, bytes):
                result[tag].append(value.decode(errors="replace"))
            elif isinstance(value, (str, bool, numbers.Number)):
                result[tag].append(value)
            elif isinstance(value, Gst.Sample):
                data = _extract_sample_data(value)
                if data:
                    result[tag].append(data)
            else:
                logger.log(
                    log.TRACE_LOG_LEVEL,
                    "Ignoring unknown tag data: %r = %r",
                    tag,
                    value,
                )

    # TODO: dict(result) to not leak the defaultdict, or just use setdefault?
    return result


def _extract_sample_data(sample):
    buf = sample.get_buffer()
    if not buf:
        return None
    return _extract_buffer_data(buf)


# Fix for https://github.com/mopidy/mopidy/issues/1827
# Using GstBuffer.extract_dup() is a memory leak in versions of PyGObject prior
# to v3.36.0. As a workaround we use the GstMemory APIs instead.
def _extract_buffer_data(buf):
    mem = buf.get_all_memory()
    if not mem:
        return None
    success, info = mem.map(Gst.MapFlags.READ)
    if not success:
        return None
    if isinstance(info.data, memoryview):
        # We need to copy the data as the memoryview is released
        # when we call mem.unmap()
        data = bytes(info.data)
    else:
        # GStreamer Python bindings <= 1.16 return a copy of the
        # data as bytes()
        data = info.data
    mem.unmap(info)
    return data


# TODO: split based on "stream" and "track" based conversion? i.e. handle data
# from radios in it's own helper instead?
def convert_tags_to_track(tags):
    """Convert our normalized tags to a track.

    :param  tags: dictionary of tag keys with a list of values
    :type tags: :class:`dict`
    :rtype: :class:`mopidy.models.Track`
    """
    album_kwargs = {}
    track_kwargs = {}

    track_kwargs["composers"] = _artists(tags, Gst.TAG_COMPOSER)
    track_kwargs["performers"] = _artists(tags, Gst.TAG_PERFORMER)
    track_kwargs["artists"] = _artists(
        tags, Gst.TAG_ARTIST, "musicbrainz-artistid", "musicbrainz-sortname"
    )
    album_kwargs["artists"] = _artists(
        tags, Gst.TAG_ALBUM_ARTIST, "musicbrainz-albumartistid"
    )

    track_kwargs["genre"] = "; ".join(tags.get(Gst.TAG_GENRE, []))
    track_kwargs["name"] = "; ".join(tags.get(Gst.TAG_TITLE, []))
    if not track_kwargs["name"]:
        track_kwargs["name"] = "; ".join(tags.get(Gst.TAG_ORGANIZATION, []))

    track_kwargs["comment"] = "; ".join(tags.get("comment", []))
    if not track_kwargs["comment"]:
        track_kwargs["comment"] = "; ".join(tags.get(Gst.TAG_LOCATION, []))
    if not track_kwargs["comment"]:
        track_kwargs["comment"] = "; ".join(tags.get(Gst.TAG_COPYRIGHT, []))

    track_kwargs["track_no"] = tags.get(Gst.TAG_TRACK_NUMBER, [None])[0]
    track_kwargs["disc_no"] = tags.get(Gst.TAG_ALBUM_VOLUME_NUMBER, [None])[0]
    track_kwargs["bitrate"] = tags.get(Gst.TAG_BITRATE, [None])[0]
    track_kwargs["musicbrainz_id"] = tags.get("musicbrainz-trackid", [None])[0]

    album_kwargs["name"] = tags.get(Gst.TAG_ALBUM, [None])[0]
    album_kwargs["num_tracks"] = tags.get(Gst.TAG_TRACK_COUNT, [None])[0]
    album_kwargs["num_discs"] = tags.get(Gst.TAG_ALBUM_VOLUME_COUNT, [None])[0]
    album_kwargs["musicbrainz_id"] = tags.get("musicbrainz-albumid", [None])[0]

    album_kwargs["date"] = tags.get(Gst.TAG_DATE, [None])[0]
    if not album_kwargs["date"]:
        datetime = tags.get(Gst.TAG_DATE_TIME, [None])[0]
        if datetime is not None:
            album_kwargs["date"] = datetime.split("T")[0]
    track_kwargs["date"] = album_kwargs["date"]

    # Clear out any empty values we found
    track_kwargs = {k: v for k, v in track_kwargs.items() if v}
    album_kwargs = {k: v for k, v in album_kwargs.items() if v}

    # Only bother with album if we have a name to show.
    if album_kwargs.get("name"):
        track_kwargs["album"] = Album(**album_kwargs)

    return Track(**track_kwargs)


def _artists(tags, artist_name, artist_id=None, artist_sortname=None):
    # Name missing, don't set artist
    if not tags.get(artist_name):
        return None

    # One artist name and either id or sortname, include all available fields
    if len(tags[artist_name]) == 1 and (
        artist_id in tags or artist_sortname in tags
    ):
        attrs = {"name": tags[artist_name][0]}
        if artist_id in tags:
            attrs["musicbrainz_id"] = tags[artist_id][0]
        if artist_sortname in tags:
            attrs["sortname"] = tags[artist_sortname][0]
        return [Artist(**attrs)]

    # Multiple artist, provide artists with name only to avoid ambiguity.
    return [Artist(name=name) for name in tags[artist_name]]



import functools
import os
import platform
import sys

import pkg_resources

from mopidy.internal import formatting
from mopidy.internal.gi import Gst, gi


def format_dependency_list(adapters=None):
    if adapters is None:
        dist_names = {
            ep.dist.project_name
            for ep in pkg_resources.iter_entry_points("mopidy.ext")
            if ep.dist.project_name != "Mopidy"
        }
        dist_infos = [
            functools.partial(pkg_info, dist_name) for dist_name in dist_names
        ]

        adapters = (
            [
                executable_info,
                platform_info,
                python_info,
                functools.partial(pkg_info, "Mopidy", True),
            ]
            + dist_infos
            + [gstreamer_info]
        )

    return "\n".join(_format_dependency(a()) for a in adapters)


def _format_dependency(dep_info):
    lines = []

    if "version" not in dep_info:
        lines.append(f"{dep_info['name']}: not found")
    else:
        source = f" from {dep_info['path']}" if "path" in dep_info else ""
        lines.append(f"{dep_info['name']}: {dep_info['version']}{source}")

    if "other" in dep_info:
        details = formatting.indent(dep_info["other"], places=4)
        lines.append(f"  Detailed information: {details}")

    if dep_info.get("dependencies", []):
        for sub_dep_info in dep_info["dependencies"]:
            sub_dep_lines = _format_dependency(sub_dep_info)
            lines.append(
                formatting.indent(sub_dep_lines, places=2, singles=True)
            )

    return "\n".join(lines)


def executable_info():
    return {
        "name": "Executable",
        "version": sys.argv[0],
    }


def platform_info():
    return {
        "name": "Platform",
        "version": platform.platform(),
    }


def python_info():
    return {
        "name": "Python",
        "version": (
            f"{platform.python_implementation()} {platform.python_version()}"
        ),
        "path": os.path.dirname(platform.__file__),
    }


def pkg_info(
    project_name=None, include_transitive_deps=True, include_extras=False
):
    if project_name is None:
        project_name = "Mopidy"
    try:
        distribution = pkg_resources.get_distribution(project_name)
        extras = include_extras and distribution.extras or []
        if include_transitive_deps:
            dependencies = [
                pkg_info(
                    d.project_name,
                    include_transitive_deps=d.project_name != "Mopidy",
                )
                for d in distribution.requires(extras)
            ]
        else:
            dependencies = []
        return {
            "name": project_name,
            "version": distribution.version,
            "path": distribution.location,
            "dependencies": dependencies,
        }
    except pkg_resources.ResolutionError:
        return {
            "name": project_name,
        }


def gstreamer_info():
    other = []
    other.append(f"Python wrapper: python-gi {gi.__version__}")

    found_elements = []
    missing_elements = []
    for name, status in _gstreamer_check_elements():
        if status:
            found_elements.append(name)
        else:
            missing_elements.append(name)

    other.append("Relevant elements:")
    other.append("  Found:")
    for element in found_elements:
        other.append(f"    {element}")
    if not found_elements:
        other.append("    none")
    other.append("  Not found:")
    for element in missing_elements:
        other.append(f"    {element}")
    if not missing_elements:
        other.append("    none")

    return {
        "name": "GStreamer",
        "version": ".".join(map(str, Gst.version())),
        "path": os.path.dirname(gi.__file__),
        "other": "\n".join(other),
    }


def _gstreamer_check_elements():
    elements_to_check = [
        # Core playback
        "uridecodebin",
        # External HTTP streams
        "souphttpsrc",
        # Spotify
        "appsrc",
        # Audio sinks
        "alsasink",
        "osssink",
        "oss4sink",
        "pulsesink",
        # MP3 encoding and decoding
        #
        # One of flump3dec, mad, and mpg123audiodec is required for MP3
        # playback.
        "flump3dec",
        "id3demux",
        "id3v2mux",
        "lamemp3enc",
        "mad",
        "mpegaudioparse",
        "mpg123audiodec",
        # Ogg Vorbis encoding and decoding
        "vorbisdec",
        "vorbisenc",
        "vorbisparse",
        "oggdemux",
        "oggmux",
        "oggparse",
        # Flac decoding
        "flacdec",
        "flacparse",
        # Shoutcast output
        "shout2send",
    ]
    known_elements = [
        factory.get_name()
        for factory in Gst.Registry.get().get_feature_list(Gst.ElementFactory)
    ]
    return [
        (element, element in known_elements) for element in elements_to_check
    ]



import os

import mopidy
from mopidy import config, ext


class Extension(ext.Extension):

    dist_name = "Mopidy-SoftwareMixer"
    ext_name = "softwaremixer"
    version = mopidy.__version__

    def get_default_config(self):
        conf_file = os.path.join(os.path.dirname(__file__), "ext.conf")
        return config.read(conf_file)

    def get_config_schema(self):
        schema = super().get_config_schema()
        return schema

    def setup(self, registry):
        from .mixer import SoftwareMixer

        registry.add("mixer", SoftwareMixer)



import unittest

from mopidy.models.fields import (
    Boolean,
    Collection,
    Field,
    Identifier,
    Integer,
    String,
)


def create_instance(field):
    """Create an instance of a dummy class for testing fields."""

    class Dummy:
        attr = field
        attr._name = "attr"

    return Dummy()


class FieldDescriptorTest(unittest.TestCase):
    def test_raw_field_accesible_through_class(self):
        field = Field()
        instance = create_instance(field)
        assert field == instance.__class__.attr

    def test_field_knows_its_name(self):
        instance = create_instance(Field())
        assert "attr" == instance.__class__.attr._name

    def test_field_has_none_as_default(self):
        instance = create_instance(Field())
        assert instance.attr is None

    def test_field_does_not_store_default(self):
        instance = create_instance(Field())
        assert not hasattr(instance, "_attr")

    def test_field_assigment_and_retrival(self):
        instance = create_instance(Field())
        instance.attr = 1234
        assert 1234 == instance.attr

    def test_field_can_be_reassigned(self):
        instance = create_instance(Field())
        instance.attr = 1234
        instance.attr = 5678
        assert 5678 == instance.attr

    def test_field_can_be_deleted(self):
        instance = create_instance(Field())
        instance.attr = 1234
        del instance.attr
        assert instance.attr is None
        assert not hasattr(instance, "_attr")

    def test_field_can_be_set_to_none(self):
        instance = create_instance(Field())
        instance.attr = 1234
        instance.attr = None
        assert instance.attr is None
        assert not hasattr(instance, "_attr")

    def test_field_can_be_set_default(self):
        default = object()
        instance = create_instance(Field(default=default))
        instance.attr = 1234
        instance.attr = default
        assert default == instance.attr
        assert not hasattr(instance, "_attr")


class FieldTest(unittest.TestCase):
    def test_default_handling(self):
        instance = create_instance(Field(default=1234))
        assert 1234 == instance.attr

    def test_type_checking(self):
        instance = create_instance(Field(type=set))
        instance.attr = set()

        with self.assertRaises(TypeError):
            instance.attr = 1234

    def test_choices_checking(self):
        instance = create_instance(Field(choices=(1, 2, 3)))
        instance.attr = 1

        with self.assertRaises(TypeError):
            instance.attr = 4

    def test_default_respects_type_check(self):
        with self.assertRaises(TypeError):
            create_instance(Field(type=int, default="123"))

    def test_default_respects_choices_check(self):
        with self.assertRaises(TypeError):
            create_instance(Field(choices=(1, 2, 3), default=5))


class StringTest(unittest.TestCase):
    def test_default_handling(self):
        instance = create_instance(String(default="abc"))
        assert "abc" == instance.attr

    def test_native_str_allowed(self):
        instance = create_instance(String())
        instance.attr = "abc"
        assert "abc" == instance.attr

    def test_unicode_allowed(self):
        instance = create_instance(String())
        instance.attr = "abc"
        assert "abc" == instance.attr

    def test_other_disallowed(self):
        instance = create_instance(String())
        with self.assertRaises(TypeError):
            instance.attr = 1234

    def test_empty_string(self):
        instance = create_instance(String())
        instance.attr = ""
        assert "" == instance.attr


class IdentifierTest(unittest.TestCase):
    def test_default_handling(self):
        instance = create_instance(Identifier(default="abc"))
        assert "abc" == instance.attr

    def test_native_str_allowed(self):
        instance = create_instance(Identifier())
        instance.attr = "abc"
        assert "abc" == instance.attr

    def test_unicode_allowed(self):
        instance = create_instance(Identifier())
        instance.attr = "abc"
        assert "abc" == instance.attr

    def test_unicode_with_nonascii_allowed(self):
        instance = create_instance(Identifier())
        instance.attr = "æøå"
        assert "æøå" == instance.attr

    def test_other_disallowed(self):
        instance = create_instance(Identifier())
        with self.assertRaises(TypeError):
            instance.attr = 1234

    def test_empty_string(self):
        instance = create_instance(Identifier())
        instance.attr = ""
        assert "" == instance.attr


class IntegerTest(unittest.TestCase):
    def test_default_handling(self):
        instance = create_instance(Integer(default=1234))
        assert 1234 == instance.attr

    def test_int_allowed(self):
        instance = create_instance(Integer())
        instance.attr = int(123)
        assert 123 == instance.attr

    def test_float_disallowed(self):
        instance = create_instance(Integer())
        with self.assertRaises(TypeError):
            instance.attr = 123.0

    def test_numeric_string_disallowed(self):
        instance = create_instance(Integer())
        with self.assertRaises(TypeError):
            instance.attr = "123"

    def test_other_disallowed(self):
        instance = create_instance(String())
        with self.assertRaises(TypeError):
            instance.attr = tuple()

    def test_min_validation(self):
        instance = create_instance(Integer(min=0))
        instance.attr = 0
        assert 0 == instance.attr

        with self.assertRaises(ValueError):
            instance.attr = -1

    def test_max_validation(self):
        instance = create_instance(Integer(max=10))
        instance.attr = 10
        assert 10 == instance.attr

        with self.assertRaises(ValueError):
            instance.attr = 11


class BooleanTest(unittest.TestCase):
    def test_default_handling(self):
        instance = create_instance(Boolean(default=True))
        assert instance.attr is True

    def test_true_allowed(self):
        instance = create_instance(Boolean())
        instance.attr = True
        assert instance.attr is True

    def test_false_allowed(self):
        instance = create_instance(Boolean())
        instance.attr = False
        assert instance.attr is False

    def test_int_forbidden(self):
        instance = create_instance(Boolean())
        with self.assertRaises(TypeError):
            instance.attr = 1


class CollectionTest(unittest.TestCase):
    def test_container_instance_is_default(self):
        instance = create_instance(Collection(type=int, container=frozenset))
        assert frozenset() == instance.attr

    def test_empty_collection(self):
        instance = create_instance(Collection(type=int, container=frozenset))
        instance.attr = []
        assert frozenset() == instance.attr

    def test_collection_gets_stored_in_container(self):
        instance = create_instance(Collection(type=int, container=frozenset))
        instance.attr = [1, 2, 3]
        assert frozenset([1, 2, 3]) == instance.attr

    def test_collection_with_wrong_type(self):
        instance = create_instance(Collection(type=int, container=frozenset))
        with self.assertRaises(TypeError):
            instance.attr = [1, "2", 3]

    def test_collection_with_string(self):
        instance = create_instance(Collection(type=int, container=frozenset))
        with self.assertRaises(TypeError):
            instance.attr = "123"

    def test_strings_should_not_be_considered_a_collection(self):
        instance = create_instance(Collection(type=str, container=tuple))
        with self.assertRaises(TypeError):
            instance.attr = b"123"



import platform

import mopidy

"Helpers for configuring HTTP clients used in Mopidy extensions."


def format_proxy(proxy_config, auth=True):
    """Convert a Mopidy proxy config to the commonly used proxy string format.

    Outputs ``scheme://host:port``, ``scheme://user:pass@host:port`` or
    :class:`None` depending on the proxy config provided.

    You can also opt out of getting the basic auth by setting ``auth`` to
    :class:`False`.

    .. versionadded:: 1.1
    """
    if not proxy_config.get("hostname"):
        return None

    scheme = proxy_config.get("scheme") or "http"
    username = proxy_config.get("username")
    password = proxy_config.get("password")
    hostname = proxy_config["hostname"]
    port = proxy_config.get("port")
    if not port or port < 0:
        port = 80

    if username and password and auth:
        return f"{scheme}://{username}:{password}@{hostname}:{port}"
    else:
        return f"{scheme}://{hostname}:{port}"


def format_user_agent(name=None):
    """Construct a User-Agent suitable for use in client code.

    This will identify use by the provided ``name`` (which should be on the
    format ``dist_name/version``), Mopidy version and Python version.

    .. versionadded:: 1.1
    """
    parts = [
        f"Mopidy/{mopidy.__version__}",
        f"{platform.python_implementation()}/{platform.python_version()}",
    ]
    if name:
        parts.insert(0, name)
    return " ".join(parts)



class PlaybackState:

    """
    Enum of playback states.
    """

    #: Constant representing the paused state.
    PAUSED = "paused"

    #: Constant representing the playing state.
    PLAYING = "playing"

    #: Constant representing the stopped state.
    STOPPED = "stopped"



import logging
import os
from unittest import mock

import pytest
import requests.exceptions
import responses

from mopidy import exceptions
from mopidy.audio import scan
from mopidy.stream import actor

TIMEOUT = 1000
PLAYLIST_URI = "http://example.com/listen.m3u"
STREAM_URI = "http://example.com/stream.mp3"
BODY = """
#EXTM3U
http://example.com/stream.mp3
http://foo.bar/baz
""".strip()


@pytest.fixture
def config():
    return {
        "proxy": {},
        "stream": {
            "timeout": TIMEOUT,
            "metadata_blacklist": [],
            "protocols": ["http"],
        },
        "file": {"enabled": False},
    }


@pytest.fixture
def audio():
    return mock.Mock()


@pytest.yield_fixture
def scanner():
    patcher = mock.patch.object(scan, "Scanner")
    yield patcher.start()()
    patcher.stop()


@pytest.fixture
def backend(audio, config, scanner):
    return actor.StreamBackend(audio=audio, config=config)


@pytest.fixture
def provider(backend):
    return backend.playback


class TestTranslateURI:
    @responses.activate
    def test_audio_stream_returns_same_uri(self, scanner, provider):
        scanner.scan.side_effect = [
            # Set playable to False to test detection by mimetype
            mock.Mock(mime="audio/mpeg", playable=False),
        ]

        result = provider.translate_uri(STREAM_URI)

        scanner.scan.assert_called_once_with(STREAM_URI, timeout=mock.ANY)
        assert result == STREAM_URI

    @responses.activate
    def test_playable_ogg_stream_is_not_considered_a_playlist(
        self, scanner, provider
    ):

        scanner.scan.side_effect = [
            # Set playable to True to ignore detection as possible playlist
            mock.Mock(mime="application/ogg", playable=True),
        ]

        result = provider.translate_uri(STREAM_URI)

        scanner.scan.assert_called_once_with(STREAM_URI, timeout=mock.ANY)
        assert result == STREAM_URI

    @responses.activate
    def test_text_playlist_with_mpeg_stream(self, scanner, provider, caplog):

        caplog.set_level(logging.DEBUG)
        scanner.scan.side_effect = [
            # Scanning playlist
            mock.Mock(mime="text/foo", playable=False),
            # Scanning stream
            mock.Mock(mime="audio/mpeg", playable=True),
        ]
        responses.add(
            responses.GET,
            PLAYLIST_URI,
            body=BODY,
            content_type="audio/x-mpegurl",
        )

        result = provider.translate_uri(PLAYLIST_URI)

        assert scanner.scan.mock_calls == [
            mock.call(PLAYLIST_URI, timeout=mock.ANY),
            mock.call(STREAM_URI, timeout=mock.ANY),
        ]
        assert result == STREAM_URI

        # Check logging to ensure debuggability
        assert f"Unwrapping stream from URI: {PLAYLIST_URI}" in caplog.text
        assert f"Parsed playlist ({PLAYLIST_URI})" in caplog.text
        assert f"Unwrapping stream from URI: {STREAM_URI}" in caplog.text
        assert (
            f"Unwrapped potential audio/mpeg stream: {STREAM_URI}"
            in caplog.text
        )

        # Check proper Requests session setup
        assert (
            responses.calls[0]
            .request.headers["User-Agent"]
            .startswith("Mopidy-Stream/")
        )

    @responses.activate
    def test_xml_playlist_with_mpeg_stream(self, scanner, provider):
        scanner.scan.side_effect = [
            # Scanning playlist
            mock.Mock(mime="application/xspf+xml", playable=False),
            # Scanning stream
            mock.Mock(mime="audio/mpeg", playable=True),
        ]
        responses.add(
            responses.GET,
            PLAYLIST_URI,
            body=BODY,
            content_type="application/xspf+xml",
        )

        result = provider.translate_uri(PLAYLIST_URI)

        assert scanner.scan.mock_calls == [
            mock.call(PLAYLIST_URI, timeout=mock.ANY),
            mock.call(STREAM_URI, timeout=mock.ANY),
        ]
        assert result == STREAM_URI

    @responses.activate
    def test_scan_fails_but_playlist_parsing_succeeds(
        self, scanner, provider, caplog
    ):

        caplog.set_level(logging.DEBUG)
        scanner.scan.side_effect = [
            # Scanning playlist
            exceptions.ScannerError("some failure"),
            # Scanning stream
            mock.Mock(mime="audio/mpeg", playable=True),
        ]
        responses.add(
            responses.GET,
            PLAYLIST_URI,
            body=BODY,
            content_type="audio/x-mpegurl",
        )

        result = provider.translate_uri(PLAYLIST_URI)

        assert f"Unwrapping stream from URI: {PLAYLIST_URI}" in caplog.text
        assert f"GStreamer failed scanning URI ({PLAYLIST_URI})" in caplog.text
        assert f"Parsed playlist ({PLAYLIST_URI})" in caplog.text
        assert (
            f"Unwrapped potential audio/mpeg stream: {STREAM_URI}"
            in caplog.text
        )
        assert result == STREAM_URI

    @responses.activate
    def test_scan_fails_and_playlist_parsing_fails(
        self, scanner, provider, caplog
    ):

        caplog.set_level(logging.DEBUG)
        scanner.scan.side_effect = exceptions.ScannerError("some failure")
        responses.add(
            responses.GET,
            STREAM_URI,
            body=b"some audio data",
            content_type="audio/mpeg",
        )

        result = provider.translate_uri(STREAM_URI)

        assert f"Unwrapping stream from URI: {STREAM_URI}" in caplog.text
        assert f"GStreamer failed scanning URI ({STREAM_URI})" in caplog.text
        assert (
            f"Failed parsing URI ({STREAM_URI}) as playlist; found potential stream."
            in caplog.text
        )
        assert result == STREAM_URI

    @responses.activate
    def test_failed_download_returns_none(self, scanner, provider, caplog):
        caplog.set_level(logging.DEBUG)
        scanner.scan.side_effect = [mock.Mock(mime="text/foo", playable=False)]

        responses.add(
            responses.GET,
            PLAYLIST_URI,
            body=requests.exceptions.HTTPError("Kaboom"),
        )

        result = provider.translate_uri(PLAYLIST_URI)

        assert result is None

        assert (
            f"Unwrapping stream from URI ({PLAYLIST_URI}) failed: "
            f"error downloading URI"
        ) in caplog.text

    @responses.activate
    def test_playlist_references_itself(self, scanner, provider, caplog):
        caplog.set_level(logging.DEBUG)
        scanner.scan.side_effect = [mock.Mock(mime="text/foo", playable=False)]
        responses.add(
            responses.GET,
            PLAYLIST_URI,
            body=BODY.replace(STREAM_URI, PLAYLIST_URI),
            content_type="audio/x-mpegurl",
        )

        result = provider.translate_uri(PLAYLIST_URI)

        assert f"Unwrapping stream from URI: {PLAYLIST_URI}" in caplog.text
        assert (
            f"Parsed playlist ({PLAYLIST_URI}) and found new URI: {PLAYLIST_URI}"
        ) in caplog.text
        assert (
            f"Unwrapping stream from URI ({PLAYLIST_URI}) failed: "
            f"playlist referenced itself"
        ) in caplog.text
        assert result is None

    @responses.activate
    def test_playlist_with_relative_mpeg_stream(
        self, scanner, provider, caplog
    ):
        caplog.set_level(logging.DEBUG)
        scanner.scan.side_effect = [
            # Scanning playlist
            mock.Mock(mime="text/foo", playable=False),
            # Scanning stream
            mock.Mock(mime="audio/mpeg", playable=True),
        ]
        responses.add(
            responses.GET,
            PLAYLIST_URI,
            body=BODY.replace(STREAM_URI, os.path.basename(STREAM_URI)),
            content_type="audio/x-mpegurl",
        )

        result = provider.translate_uri(PLAYLIST_URI)

        assert scanner.scan.mock_calls == [
            mock.call(PLAYLIST_URI, timeout=mock.ANY),
            mock.call(STREAM_URI, timeout=mock.ANY),
        ]
        assert result == STREAM_URI

        assert (
            f"Parsed playlist ({PLAYLIST_URI}) and found new URI: "
            f"{os.path.basename(STREAM_URI)}"
        ) in caplog.text
        assert f"Unwrapping stream from URI: {STREAM_URI}" in caplog.text



import sys
import textwrap

try:
    import gi

    gi.require_version("Gst", "1.0")
    from gi.repository import GLib, GObject, Gst
except ImportError:
    print(
        textwrap.dedent(
            """
        ERROR: A GObject based library was not found.

        Mopidy requires GStreamer to work. GStreamer is a C library with a
        number of dependencies itself, and cannot be installed with the regular
        Python tools like pip.

        Please see https://docs.mopidy.com/en/latest/installation/ for
        instructions on how to install the required dependencies.
    """
        )
    )
    raise
else:
    Gst.init([])
    gi.require_version("GstPbutils", "1.0")
    from gi.repository import GstPbutils

GLib.set_prgname("mopidy")
GLib.set_application_name("Mopidy")

REQUIRED_GST_VERSION = (1, 14, 0)
REQUIRED_GST_VERSION_DISPLAY = ".".join(map(str, REQUIRED_GST_VERSION))

if Gst.version() < REQUIRED_GST_VERSION:
    sys.exit(
        f"ERROR: Mopidy requires GStreamer >= {REQUIRED_GST_VERSION_DISPLAY}, "
        f"but found {Gst.version_string()}."
    )


__all__ = [
    "GLib",
    "GObject",
    "Gst",
    "GstPbutils",
    "gi",
]



import collections
import logging
import time

from mopidy import exceptions
from mopidy.audio import tags as tags_lib
from mopidy.audio import utils
from mopidy.internal import log
from mopidy.internal.gi import Gst, GstPbutils

# GST_ELEMENT_FACTORY_LIST:
_DECODER = 1 << 0
_AUDIO = 1 << 50
_DEMUXER = 1 << 5
_DEPAYLOADER = 1 << 8
_PARSER = 1 << 6

# GST_TYPE_AUTOPLUG_SELECT_RESULT:
_SELECT_TRY = 0
_SELECT_EXPOSE = 1

_Result = collections.namedtuple(
    "Result", ("uri", "tags", "duration", "seekable", "mime", "playable")
)

logger = logging.getLogger(__name__)


def _trace(*args, **kwargs):
    logger.log(log.TRACE_LOG_LEVEL, *args, **kwargs)


# TODO: replace with a scan(uri, timeout=1000, proxy_config=None)?
class Scanner:

    """
    Helper to get tags and other relevant info from URIs.

    :param timeout: timeout for scanning a URI in ms
    :param proxy_config: dictionary containing proxy config strings.
    :type event: int
    """

    def __init__(self, timeout=1000, proxy_config=None):
        self._timeout_ms = int(timeout)
        self._proxy_config = proxy_config or {}

    def scan(self, uri, timeout=None):
        """
        Scan the given uri collecting relevant metadata.

        :param uri: URI of the resource to scan.
        :type uri: string
        :param timeout: timeout for scanning a URI in ms. Defaults to the
            ``timeout`` value used when creating the scanner.
        :type timeout: int
        :return: A named tuple containing
            ``(uri, tags, duration, seekable, mime)``.
            ``tags`` is a dictionary of lists for all the tags we found.
            ``duration`` is the length of the URI in milliseconds, or
            :class:`None` if the URI has no duration. ``seekable`` is boolean.
            indicating if a seek would succeed.
        """
        timeout = int(timeout or self._timeout_ms)
        tags, duration, seekable, mime = None, None, None, None
        pipeline, signals = _setup_pipeline(uri, self._proxy_config)

        try:
            _start_pipeline(pipeline)
            tags, mime, have_audio, duration = _process(pipeline, timeout)
            seekable = _query_seekable(pipeline)
        finally:
            signals.clear()
            pipeline.set_state(Gst.State.NULL)
            del pipeline

        return _Result(uri, tags, duration, seekable, mime, have_audio)


# Turns out it's _much_ faster to just create a new pipeline for every as
# decodebins and other elements don't seem to take well to being reused.
def _setup_pipeline(uri, proxy_config=None):
    src = Gst.Element.make_from_uri(Gst.URIType.SRC, uri)
    if not src:
        raise exceptions.ScannerError(f"GStreamer can not open: {uri}")

    if proxy_config:
        utils.setup_proxy(src, proxy_config)

    signals = utils.Signals()
    pipeline = Gst.ElementFactory.make("pipeline")
    pipeline.add(src)

    if _has_src_pads(src):
        _setup_decodebin(src, src.get_static_pad("src"), pipeline, signals)
    elif _has_dynamic_src_pad(src):
        signals.connect(src, "pad-added", _setup_decodebin, pipeline, signals)
    else:
        raise exceptions.ScannerError("No pads found in source element.")

    return pipeline, signals


def _has_src_pads(element):
    pads = []
    element.iterate_src_pads().foreach(pads.append)
    return bool(pads)


def _has_dynamic_src_pad(element):
    for template in element.get_pad_template_list():
        if template.direction == Gst.PadDirection.SRC:
            if template.presence == Gst.PadPresence.SOMETIMES:
                return True
    return False


def _setup_decodebin(element, pad, pipeline, signals):
    typefind = Gst.ElementFactory.make("typefind")
    decodebin = Gst.ElementFactory.make("decodebin")

    for element in (typefind, decodebin):
        pipeline.add(element)
        element.sync_state_with_parent()

    pad.link(typefind.get_static_pad("sink"))
    typefind.link(decodebin)

    signals.connect(typefind, "have-type", _have_type, decodebin)
    signals.connect(decodebin, "pad-added", _pad_added, pipeline)
    signals.connect(decodebin, "autoplug-select", _autoplug_select)


def _have_type(element, probability, caps, decodebin):
    decodebin.set_property("sink-caps", caps)
    struct = Gst.Structure.new_empty("have-type")
    struct.set_value("caps", caps.get_structure(0))
    element.get_bus().post(Gst.Message.new_application(element, struct))


def _pad_added(element, pad, pipeline):
    sink = Gst.ElementFactory.make("fakesink")
    sink.set_property("sync", False)

    pipeline.add(sink)
    sink.sync_state_with_parent()
    pad.link(sink.get_static_pad("sink"))

    if pad.query_caps().is_subset(Gst.Caps.from_string("audio/x-raw")):
        # Probably won't happen due to autoplug-select fix, but lets play it
        # safe until we've tested more.
        struct = Gst.Structure.new_empty("have-audio")
        element.get_bus().post(Gst.Message.new_application(element, struct))


def _autoplug_select(element, pad, caps, factory):
    if factory.list_is_type(_DECODER | _AUDIO):
        struct = Gst.Structure.new_empty("have-audio")
        element.get_bus().post(Gst.Message.new_application(element, struct))
    if not factory.list_is_type(_DEMUXER | _DEPAYLOADER | _PARSER):
        return _SELECT_EXPOSE
    return _SELECT_TRY


def _start_pipeline(pipeline):
    result = pipeline.set_state(Gst.State.PAUSED)
    if result == Gst.StateChangeReturn.NO_PREROLL:
        pipeline.set_state(Gst.State.PLAYING)


def _query_duration(pipeline):
    success, duration = pipeline.query_duration(Gst.Format.TIME)
    if not success:
        duration = None  # Make sure error case preserves None.
    elif duration < 0:
        duration = None  # Stream without duration.
    else:
        duration = int(duration // Gst.MSECOND)
    return success, duration


def _query_seekable(pipeline):
    query = Gst.Query.new_seeking(Gst.Format.TIME)
    pipeline.query(query)
    return query.parse_seeking()[1]


def _process(pipeline, timeout_ms):
    bus = pipeline.get_bus()
    tags = {}
    mime = None
    have_audio = False
    missing_message = None
    duration = None

    types = (
        Gst.MessageType.ELEMENT
        | Gst.MessageType.APPLICATION
        | Gst.MessageType.ERROR
        | Gst.MessageType.EOS
        | Gst.MessageType.ASYNC_DONE
        | Gst.MessageType.DURATION_CHANGED
        | Gst.MessageType.TAG
    )

    timeout = timeout_ms
    start = int(time.time() * 1000)
    while timeout > 0:
        msg = bus.timed_pop_filtered(timeout * Gst.MSECOND, types)
        if msg is None:
            break

        if logger.isEnabledFor(log.TRACE_LOG_LEVEL) and msg.get_structure():
            debug_text = msg.get_structure().to_string()
            if len(debug_text) > 77:
                debug_text = debug_text[:77] + "..."
            _trace("element %s: %s", msg.src.get_name(), debug_text)

        if msg.type == Gst.MessageType.ELEMENT:
            if GstPbutils.is_missing_plugin_message(msg):
                missing_message = msg
        elif msg.type == Gst.MessageType.APPLICATION:
            if msg.get_structure().get_name() == "have-type":
                mime = msg.get_structure().get_value("caps").get_name()
                if mime and (
                    mime.startswith("text/") or mime == "application/xml"
                ):
                    return tags, mime, have_audio, duration
            elif msg.get_structure().get_name() == "have-audio":
                have_audio = True
        elif msg.type == Gst.MessageType.ERROR:
            error, _debug = msg.parse_error()
            if missing_message and not mime:
                caps = missing_message.get_structure().get_value("detail")
                mime = caps.get_structure(0).get_name()
                return tags, mime, have_audio, duration
            raise exceptions.ScannerError(str(error))
        elif msg.type == Gst.MessageType.EOS:
            return tags, mime, have_audio, duration
        elif msg.type == Gst.MessageType.ASYNC_DONE:
            success, duration = _query_duration(pipeline)
            if tags and success:
                return tags, mime, have_audio, duration

            # Don't try workaround for non-seekable sources such as mmssrc:
            if not _query_seekable(pipeline):
                return tags, mime, have_audio, duration

            # Workaround for upstream bug which causes tags/duration to arrive
            # after pre-roll. We get around this by starting to play the track
            # and then waiting for a duration change.
            # https://bugzilla.gnome.org/show_bug.cgi?id=763553
            logger.debug("Using workaround for duration missing before play.")
            result = pipeline.set_state(Gst.State.PLAYING)
            if result == Gst.StateChangeReturn.FAILURE:
                return tags, mime, have_audio, duration

        elif msg.type == Gst.MessageType.DURATION_CHANGED and tags:
            # VBR formats sometimes seem to not have a duration by the time we
            # go back to paused. So just try to get it right away.
            success, duration = _query_duration(pipeline)
            pipeline.set_state(Gst.State.PAUSED)
            if success:
                return tags, mime, have_audio, duration
        elif msg.type == Gst.MessageType.TAG:
            taglist = msg.parse_tag()
            # Note that this will only keep the last tag.
            tags.update(tags_lib.convert_taglist(taglist))

        timeout = timeout_ms - (int(time.time() * 1000) - start)

    raise exceptions.ScannerError(f"Timeout after {timeout_ms:d}ms")


if __name__ == "__main__":
    import os
    import sys

    from mopidy.internal import path

    logging.basicConfig(
        format="%(asctime)-15s %(levelname)s %(message)s",
        level=log.TRACE_LOG_LEVEL,
    )

    scanner = Scanner(5000)
    for uri in sys.argv[1:]:
        if not Gst.uri_is_valid(uri):
            uri = path.path_to_uri(os.path.abspath(uri))
        try:
            result = scanner.scan(uri)
            for key in ("uri", "mime", "duration", "playable", "seekable"):
                value = getattr(result, key)
                print(f"{key:<20}   {value}")
            print("tags")
            for tag, value in result.tags.items():
                line = f"{tag:<20}   {value}"
                if len(line) > 77:
                    line = line[:77] + "..."
                print(line)
        except exceptions.ScannerError as error:
            print(f"{uri}: {error}")



import os
import shutil
import tempfile
import urllib
from unittest import mock

import tornado.testing
import tornado.wsgi

import mopidy
from mopidy.http import actor, handlers


class HttpServerTest(tornado.testing.AsyncHTTPTestCase):
    def get_config(self):
        return {
            "http": {
                "hostname": "127.0.0.1",
                "port": 6680,
                "zeroconf": "",
                "allowed_origins": [],
                "csrf_protection": True,
                "default_app": "mopidy",
            }
        }

    def get_app(self):
        core = mock.Mock()
        core.get_version = mock.MagicMock(name="get_version")
        core.get_version.return_value = mopidy.__version__

        testapps = [dict(name="testapp")]
        teststatics = [dict(name="teststatic")]

        apps = [
            {
                "name": "mopidy",
                "factory": handlers.make_mopidy_app_factory(
                    testapps, teststatics
                ),
            }
        ]

        http_server = actor.HttpServer(
            config=self.get_config(),
            core=core,
            sockets=[],
            apps=apps,
            statics=[],
        )

        return tornado.web.Application(http_server._get_request_handlers())


class RootRedirectTest(HttpServerTest):
    def test_should_redirect_to_mopidy_app(self):
        response = self.fetch("/", method="GET", follow_redirects=False)

        assert response.code == 302
        assert response.headers["Location"] == "/mopidy/"


class MopidyAppTest(HttpServerTest):
    def test_should_return_index(self):
        response = self.fetch("/mopidy/", method="GET")
        body = response.body.decode()

        assert "This web server is a part of the Mopidy music server." in body
        assert "testapp" in body
        assert "teststatic" in body
        assert response.headers["X-Mopidy-Version"] == mopidy.__version__
        assert response.headers["Cache-Control"] == "no-cache"

    def test_without_slash_should_redirect(self):
        response = self.fetch("/mopidy", method="GET", follow_redirects=False)

        assert response.code == 301
        assert response.headers["Location"] == "/mopidy/"

    def test_should_return_static_files(self):
        response = self.fetch("/mopidy/mopidy.css", method="GET")

        assert "html {" in response.body.decode()
        assert response.headers["X-Mopidy-Version"] == mopidy.__version__
        assert response.headers["Cache-Control"] == "no-cache"


class MopidyWebSocketHandlerTest(HttpServerTest):
    def test_should_return_ws(self):
        response = self.fetch("/mopidy/ws", method="GET")

        assert 'Can "Upgrade" only to "WebSocket".' == response.body.decode()

    def test_should_return_ws_old(self):
        response = self.fetch("/mopidy/ws/", method="GET")

        assert 'Can "Upgrade" only to "WebSocket".' == response.body.decode()


class MopidyRPCHandlerTest(HttpServerTest):
    def test_should_return_rpc_error(self):
        cmd = tornado.escape.json_encode({"action": "get_version"})

        response = self.fetch(
            "/mopidy/rpc",
            method="POST",
            body=cmd,
            headers={"Content-Type": "application/json"},
        )

        assert {
            "jsonrpc": "2.0",
            "id": None,
            "error": {
                "message": "Invalid Request",
                "code": (-32600),
                "data": "'jsonrpc' member must be included",
            },
        } == tornado.escape.json_decode(response.body)

    def test_should_return_parse_error(self):
        cmd = "{[[[]}"

        response = self.fetch(
            "/mopidy/rpc",
            method="POST",
            body=cmd,
            headers={"Content-Type": "application/json"},
        )

        assert {
            "jsonrpc": "2.0",
            "id": None,
            "error": {"message": "Parse error", "code": (-32700)},
        } == tornado.escape.json_decode(response.body)

    def test_should_return_mopidy_version(self):
        cmd = tornado.escape.json_encode(
            {
                "method": "core.get_version",
                "params": [],
                "jsonrpc": "2.0",
                "id": 1,
            }
        )

        response = self.fetch(
            "/mopidy/rpc",
            method="POST",
            body=cmd,
            headers={"Content-Type": "application/json"},
        )

        assert {
            "jsonrpc": "2.0",
            "id": 1,
            "result": mopidy.__version__,
        } == tornado.escape.json_decode(response.body)

    def test_should_return_extra_headers(self):
        response = self.fetch("/mopidy/rpc", method="HEAD")

        assert "Accept" in response.headers
        assert "X-Mopidy-Version" in response.headers
        assert "Cache-Control" in response.headers
        assert "Content-Type" in response.headers

    def test_should_require_correct_content_type(self):
        cmd = tornado.escape.json_encode(
            {
                "method": "core.get_version",
                "params": [],
                "jsonrpc": "2.0",
                "id": 1,
            }
        )

        response = self.fetch(
            "/mopidy/rpc",
            method="POST",
            body=cmd,
            headers={"Content-Type": "text/plain"},
        )

        assert response.code == 415
        assert response.reason == "Content-Type must be application/json"

    def test_different_origin_returns_access_denied(self):
        response = self.fetch(
            "/mopidy/rpc",
            method="OPTIONS",
            headers={"Host": "me:6680", "Origin": "http://evil:666"},
        )

        assert response.code == 403
        assert response.reason == "Access denied for origin http://evil:666"

    def test_same_origin_returns_cors_headers(self):
        response = self.fetch(
            "/mopidy/rpc",
            method="OPTIONS",
            headers={"Host": "me:6680", "Origin": "http://me:6680"},
        )

        assert (
            response.headers["Access-Control-Allow-Origin"] == "http://me:6680"
        )
        assert (
            response.headers["Access-Control-Allow-Headers"] == "Content-Type"
        )


class MopidyRPCHandlerNoCSRFProtectionTest(HttpServerTest):
    def get_config(self):
        config = super().get_config()
        config["http"]["csrf_protection"] = False
        return config

    def get_cmd(self):
        return tornado.escape.json_encode(
            {
                "method": "core.get_version",
                "params": [],
                "jsonrpc": "2.0",
                "id": 1,
            }
        )

    def test_should_ignore_incorrect_content_type(self):
        response = self.fetch(
            "/mopidy/rpc",
            method="POST",
            body=self.get_cmd(),
            headers={"Content-Type": "text/plain"},
        )

        assert response.code == 200

    def test_should_ignore_missing_content_type(self):
        response = self.fetch(
            "/mopidy/rpc", method="POST", body=self.get_cmd(), headers={}
        )

        assert response.code == 200

    def test_different_origin_returns_allowed(self):
        response = self.fetch(
            "/mopidy/rpc",
            method="OPTIONS",
            headers={"Host": "me:6680", "Origin": "http://evil:666"},
        )

        assert response.code == 204

    def test_should_not_return_cors_headers(self):
        response = self.fetch(
            "/mopidy/rpc",
            method="OPTIONS",
            headers={"Host": "me:6680", "Origin": "http://me:6680"},
        )

        assert "Access-Control-Allow-Origin" not in response.headers
        assert "Access-Control-Allow-Headers" not in response.headers


class HttpServerWithStaticFilesTest(tornado.testing.AsyncHTTPTestCase):
    def get_app(self):
        config = {
            "http": {
                "hostname": "127.0.0.1",
                "port": 6680,
                "zeroconf": "",
                "default_app": "static",
            }
        }
        core = mock.Mock()

        statics = [dict(name="static", path=os.path.dirname(__file__))]

        http_server = actor.HttpServer(
            config=config, core=core, sockets=[], apps=[], statics=statics
        )

        return tornado.web.Application(http_server._get_request_handlers())

    def test_without_slash_should_redirect(self):
        response = self.fetch("/static", method="GET", follow_redirects=False)

        assert response.code == 301
        assert response.headers["Location"] == "/static/"

    def test_can_serve_static_files(self):
        response = self.fetch("/static/test_server.py", method="GET")

        assert 200 == response.code
        assert response.headers["X-Mopidy-Version"] == mopidy.__version__
        assert response.headers["Cache-Control"] == "no-cache"


def wsgi_app_factory(config, core):
    def wsgi_app(environ, start_response):
        status = "200 OK"
        response_headers = [("Content-type", "text/plain")]
        start_response(status, response_headers)
        return [b"Hello, world!\n"]

    return [
        (
            "(.*)",
            tornado.web.FallbackHandler,
            {"fallback": tornado.wsgi.WSGIContainer(wsgi_app)},
        ),
    ]


class HttpServerWithWsgiAppTest(tornado.testing.AsyncHTTPTestCase):
    def get_app(self):
        config = {
            "http": {
                "hostname": "127.0.0.1",
                "port": 6680,
                "zeroconf": "",
                "default_app": "wsgi",
            }
        }
        core = mock.Mock()

        apps = [{"name": "wsgi", "factory": wsgi_app_factory}]

        http_server = actor.HttpServer(
            config=config, core=core, sockets=[], apps=apps, statics=[]
        )

        return tornado.web.Application(http_server._get_request_handlers())

    def test_without_slash_should_redirect(self):
        response = self.fetch("/wsgi", method="GET", follow_redirects=False)

        assert response.code == 301
        assert response.headers["Location"] == "/wsgi/"

    def test_can_wrap_wsgi_apps(self):
        response = self.fetch("/wsgi/", method="GET")

        assert response.code == 200
        assert "Hello, world!" in response.body.decode()


def default_webapp_factory(config, core):
    class MainHandler(tornado.web.RequestHandler):
        def get(self):
            self.write("Hello from default webapp")

    return [("/", MainHandler, {})]


class HttpServerWithAppDefaultApp(tornado.testing.AsyncHTTPTestCase):
    def get_app(self):
        config = {
            "http": {
                "hostname": "127.0.0.1",
                "port": 6680,
                "zeroconf": "",
                "default_app": "default_app",
            }
        }
        core = mock.Mock()

        apps = [dict(name="default_app", factory=default_webapp_factory)]

        http_server = actor.HttpServer(
            config=config, core=core, sockets=[], apps=apps, statics=[]
        )

        return tornado.web.Application(http_server._get_request_handlers())

    def test_should_redirect_to_default_app(self):
        response = self.fetch("/", method="GET", follow_redirects=False)

        assert response.code == 302
        assert response.headers["Location"] == "/default_app/"

        response = self.fetch(
            "/default_app/", method="GET", follow_redirects=True
        )

        assert response.code == 200
        assert "Hello from default webapp" in response.body.decode()


class HttpServerWithStaticDefaultApp(tornado.testing.AsyncHTTPTestCase):
    def get_app(self):
        config = {
            "http": {
                "hostname": "127.0.0.1",
                "port": 6680,
                "zeroconf": "",
                "default_app": "default_app",
            }
        }
        core = mock.Mock()

        statics = [dict(name="default_app", path=os.path.dirname(__file__))]

        http_server = actor.HttpServer(
            config=config, core=core, sockets=[], apps=[], statics=statics
        )

        return tornado.web.Application(http_server._get_request_handlers())

    def test_should_redirect_to_default_app(self):
        response = self.fetch("/", method="GET", follow_redirects=False)

        assert response.code == 302
        assert response.headers["Location"] == "/default_app/"


class HttpServerWithInvalidDefaultApp(HttpServerTest):
    def get_config(self):
        config = super(HttpServerWithInvalidDefaultApp, self).get_config()
        config["http"]["default_app"] = "invalid_webclient"
        return config

    def test_should_redirect_to_clients_list(self):
        response = self.fetch("/", method="GET", follow_redirects=False)

        assert response.code == 302
        assert response.headers["Location"] == "/mopidy/"

        response = self.fetch("/", method="GET")
        body = response.body.decode()

        assert "This web server is a part of the Mopidy music server." in body

        assert "testapp" in body
        assert "teststatic" in body
        assert response.headers["X-Mopidy-Version"] == mopidy.__version__
        assert response.headers["Cache-Control"] == "no-cache"


def cookie_secret_app_factory(config, core):
    class BaseHandler(tornado.web.RequestHandler):
        def get_current_user(self):
            return self.get_secure_cookie("user")

    class LoginHandler(BaseHandler):
        def get(self):
            self.write("This is a login form")

        def post(self):
            self.set_secure_cookie("user", self.get_argument("name"))
            self.write("Logged in")

    class MainHandler(BaseHandler):
        def get(self):
            if not self.current_user:
                self.write("Unknown user...")
                return

            name = tornado.escape.xhtml_escape(self.current_user)
            self.write("Hello, " + name)

    return [("/", MainHandler, {}), ("/login", LoginHandler, {})]


class HttpServerTestLoginWithSecureCookie(tornado.testing.AsyncHTTPTestCase):
    def get_app(self):
        self._dirpath = tempfile.mkdtemp()

        config = {
            "http": {
                "hostname": "127.0.0.1",
                "port": 6680,
                "zeroconf": "",
                "default_app": "mopidy",
            },
            "core": {"data_dir": self._dirpath},
        }
        core = mock.Mock()

        apps = [{"name": "cookie_secret", "factory": cookie_secret_app_factory}]

        http_server = actor.HttpServer(
            config=config, core=core, sockets=[], apps=apps, statics=[]
        )

        return tornado.web.Application(
            http_server._get_request_handlers(),
            cookie_secret=http_server._get_cookie_secret(),
        )

    def test_main_access_without_login(self):
        response = self.fetch("/cookie_secret", method="GET")

        assert 200 == response.code
        assert "Unknown user..." in response.body.decode()

    def test_accessing_login_form_get(self):
        response = self.fetch("/cookie_secret/login", method="GET")

        assert 200 == response.code
        assert "This is a login form" in response.body.decode()

    def test_login(self):
        post_data = {"name": "theuser"}
        body = urllib.parse.urlencode(post_data)

        response = self.fetch("/cookie_secret/login", method="POST", body=body)

        assert 200 == response.code
        assert "Logged in" in response.body.decode()

        shutil.rmtree(self._dirpath)


def test_get_secure_cookie(tmp_path):
    config = {
        "http": {
            "hostname": "127.0.0.1",
            "port": 6680,
            "zeroconf": "",
            "default_app": "mopidy",
        },
        "core": {"data_dir": tmp_path},
    }
    core = mock.Mock()

    http_server = actor.HttpServer(
        config=config, core=core, sockets=[], apps=[], statics=[]
    )

    # first secret, generating
    secret_1 = http_server._get_cookie_secret()

    assert isinstance(secret_1, str)
    assert secret_1 != ""
    assert len(secret_1) == 64

    # second secret, from file
    secret_2 = http_server._get_cookie_secret()
    assert secret_1 == secret_2



import unittest

from mopidy.audio import tags
from mopidy.internal.gi import GLib, GObject, Gst
from mopidy.models import Album, Artist, Track


class TestReprTags:
    def test_bytes_truncated_default(self):
        taglist = {"foo": [b"abcdefghijkl", b"mnop", "qrstuvwxyzabc"]}

        result = tags.repr_tags(taglist)

        assert result == "{'foo': [b'abcdefghij...', b'mnop', 'qrstuvwxyzabc']}"

    def test_max_bytes_two(self):
        taglist = {"foo": [b"1234", b"5678", "abcd", 67]}

        result = tags.repr_tags(taglist, 2)

        assert result == "{'foo': [b'12...', b'56...', 'abcd', 67]}"


class TestConvertTaglist:
    def make_taglist(self, tag, values):
        taglist = Gst.TagList.new_empty()

        for value in values:
            if isinstance(value, (GLib.Date, Gst.DateTime)):
                taglist.add_value(Gst.TagMergeMode.APPEND, tag, value)
                continue

            gobject_value = GObject.Value()
            if isinstance(value, bytes):
                gobject_value.init(GObject.TYPE_STRING)
                gobject_value.set_string(value.decode())
            elif isinstance(value, str):
                gobject_value.init(GObject.TYPE_STRING)
                gobject_value.set_string(value)
            elif isinstance(value, int):
                gobject_value.init(GObject.TYPE_UINT)
                gobject_value.set_uint(value)
            else:
                raise TypeError
            taglist.add_value(Gst.TagMergeMode.APPEND, tag, gobject_value)

        return taglist

    def test_date_tag(self):
        date = GLib.Date.new_dmy(7, 1, 2014)
        taglist = self.make_taglist(Gst.TAG_DATE, [date])

        result = tags.convert_taglist(taglist)

        assert isinstance(result[Gst.TAG_DATE][0], str)
        assert result[Gst.TAG_DATE][0] == "2014-01-07"

    def test_date_tag_bad_value(self):
        date = GLib.Date.new_dmy(7, 1, 10000)
        taglist = self.make_taglist(Gst.TAG_DATE, [date])

        result = tags.convert_taglist(taglist)

        assert len(result[Gst.TAG_DATE]) == 0

    def test_date_time_tag(self):
        taglist = self.make_taglist(
            Gst.TAG_DATE_TIME,
            [Gst.DateTime.new_from_iso8601_string("2014-01-07 14:13:12")],
        )

        result = tags.convert_taglist(taglist)

        assert isinstance(result[Gst.TAG_DATE_TIME][0], str)
        assert result[Gst.TAG_DATE_TIME][0] == "2014-01-07T14:13:12Z"

    def test_string_tag(self):
        taglist = self.make_taglist(Gst.TAG_ARTIST, [b"ABBA", b"ACDC"])

        result = tags.convert_taglist(taglist)

        assert isinstance(result[Gst.TAG_ARTIST][0], str)
        assert result[Gst.TAG_ARTIST][0] == "ABBA"
        assert isinstance(result[Gst.TAG_ARTIST][1], str)
        assert result[Gst.TAG_ARTIST][1] == "ACDC"

    def test_integer_tag(self):
        taglist = self.make_taglist(Gst.TAG_BITRATE, [17])

        result = tags.convert_taglist(taglist)

        assert result[Gst.TAG_BITRATE][0] == 17


# TODO: keep ids without name?
# TODO: current test is trying to test everything at once with a complete tags
# set, instead we might want to try with a minimal one making testing easier.
class TagsToTrackTest(unittest.TestCase):
    def setUp(self):  # noqa: N802
        self.tags = {
            "album": ["album"],
            "track-number": [1],
            "artist": ["artist"],
            "composer": ["composer"],
            "performer": ["performer"],
            "album-artist": ["albumartist"],
            "title": ["track"],
            "track-count": [2],
            "album-disc-number": [2],
            "album-disc-count": [3],
            "date": ["2006-01-01"],
            "container-format": ["ID3 tag"],
            "genre": ["genre"],
            "comment": ["comment"],
            "musicbrainz-trackid": ["trackid"],
            "musicbrainz-albumid": ["albumid"],
            "musicbrainz-artistid": ["artistid"],
            "musicbrainz-sortname": ["sortname"],
            "musicbrainz-albumartistid": ["albumartistid"],
            "bitrate": [1000],
        }

        artist = Artist(
            name="artist", musicbrainz_id="artistid", sortname="sortname"
        )
        composer = Artist(name="composer")
        performer = Artist(name="performer")
        albumartist = Artist(name="albumartist", musicbrainz_id="albumartistid")

        album = Album(
            name="album",
            date="2006-01-01",
            num_tracks=2,
            num_discs=3,
            musicbrainz_id="albumid",
            artists=[albumartist],
        )

        self.track = Track(
            name="track",
            date="2006-01-01",
            genre="genre",
            track_no=1,
            disc_no=2,
            comment="comment",
            musicbrainz_id="trackid",
            album=album,
            bitrate=1000,
            artists=[artist],
            composers=[composer],
            performers=[performer],
        )

    def check(self, expected):
        actual = tags.convert_tags_to_track(self.tags)
        assert expected == actual

    def test_track(self):
        self.check(self.track)

    def test_missing_track_no(self):
        del self.tags["track-number"]
        self.check(self.track.replace(track_no=None))

    def test_multiple_track_no(self):
        self.tags["track-number"].append(9)
        self.check(self.track)

    def test_missing_track_disc_no(self):
        del self.tags["album-disc-number"]
        self.check(self.track.replace(disc_no=None))

    def test_multiple_track_disc_no(self):
        self.tags["album-disc-number"].append(9)
        self.check(self.track)

    def test_missing_track_name(self):
        del self.tags["title"]
        self.check(self.track.replace(name=None))

    def test_multiple_track_name(self):
        self.tags["title"] = ["name1", "name2"]
        self.check(self.track.replace(name="name1; name2"))

    def test_missing_track_musicbrainz_id(self):
        del self.tags["musicbrainz-trackid"]
        self.check(self.track.replace(musicbrainz_id=None))

    def test_multiple_track_musicbrainz_id(self):
        self.tags["musicbrainz-trackid"].append("id")
        self.check(self.track)

    def test_missing_track_bitrate(self):
        del self.tags["bitrate"]
        self.check(self.track.replace(bitrate=None))

    def test_multiple_track_bitrate(self):
        self.tags["bitrate"].append(1234)
        self.check(self.track)

    def test_missing_track_genre(self):
        del self.tags["genre"]
        self.check(self.track.replace(genre=None))

    def test_multiple_track_genre(self):
        self.tags["genre"] = ["genre1", "genre2"]
        self.check(self.track.replace(genre="genre1; genre2"))

    def test_missing_track_date(self):
        del self.tags["date"]
        self.check(
            self.track.replace(
                album=self.track.album.replace(date=None), date=None
            )
        )

    def test_multiple_track_date(self):
        self.tags["date"].append("2030-01-01")
        self.check(self.track)

    def test_datetime_instead_of_date(self):
        del self.tags["date"]
        self.tags["datetime"] = ["2006-01-01T14:13:12Z"]
        self.check(self.track)

    def test_missing_track_comment(self):
        del self.tags["comment"]
        self.check(self.track.replace(comment=None))

    def test_multiple_track_comment(self):
        self.tags["comment"] = ["comment1", "comment2"]
        self.check(self.track.replace(comment="comment1; comment2"))

    def test_missing_track_artist_name(self):
        del self.tags["artist"]
        self.check(self.track.replace(artists=[]))

    def test_multiple_track_artist_name(self):
        self.tags["artist"] = ["name1", "name2"]
        artists = [Artist(name="name1"), Artist(name="name2")]
        self.check(self.track.replace(artists=artists))

    def test_missing_track_artist_musicbrainz_id(self):
        del self.tags["musicbrainz-artistid"]
        artist = list(self.track.artists)[0].replace(musicbrainz_id=None)
        self.check(self.track.replace(artists=[artist]))

    def test_multiple_track_artist_musicbrainz_id(self):
        self.tags["musicbrainz-artistid"].append("id")
        self.check(self.track)

    def test_missing_track_composer_name(self):
        del self.tags["composer"]
        self.check(self.track.replace(composers=[]))

    def test_multiple_track_composer_name(self):
        self.tags["composer"] = ["composer1", "composer2"]
        composers = [Artist(name="composer1"), Artist(name="composer2")]
        self.check(self.track.replace(composers=composers))

    def test_missing_track_performer_name(self):
        del self.tags["performer"]
        self.check(self.track.replace(performers=[]))

    def test_multiple_track_performe_name(self):
        self.tags["performer"] = ["performer1", "performer2"]
        performers = [Artist(name="performer1"), Artist(name="performer2")]
        self.check(self.track.replace(performers=performers))

    def test_missing_album_name(self):
        del self.tags["album"]
        self.check(self.track.replace(album=None))

    def test_multiple_album_name(self):
        self.tags["album"].append("album2")
        self.check(self.track)

    def test_missing_album_musicbrainz_id(self):
        del self.tags["musicbrainz-albumid"]
        album = self.track.album.replace(musicbrainz_id=None)
        self.check(self.track.replace(album=album))

    def test_multiple_album_musicbrainz_id(self):
        self.tags["musicbrainz-albumid"].append("id")
        self.check(self.track)

    def test_missing_album_num_tracks(self):
        del self.tags["track-count"]
        album = self.track.album.replace(num_tracks=None)
        self.check(self.track.replace(album=album))

    def test_multiple_album_num_tracks(self):
        self.tags["track-count"].append(9)
        self.check(self.track)

    def test_missing_album_num_discs(self):
        del self.tags["album-disc-count"]
        album = self.track.album.replace(num_discs=None)
        self.check(self.track.replace(album=album))

    def test_multiple_album_num_discs(self):
        self.tags["album-disc-count"].append(9)
        self.check(self.track)

    def test_missing_album_artist_name(self):
        del self.tags["album-artist"]
        album = self.track.album.replace(artists=[])
        self.check(self.track.replace(album=album))

    def test_multiple_album_artist_name(self):
        self.tags["album-artist"] = ["name1", "name2"]
        artists = [Artist(name="name1"), Artist(name="name2")]
        album = self.track.album.replace(artists=artists)
        self.check(self.track.replace(album=album))

    def test_missing_album_artist_musicbrainz_id(self):
        del self.tags["musicbrainz-albumartistid"]
        albumartist = list(self.track.album.artists)[0]
        albumartist = albumartist.replace(musicbrainz_id=None)
        album = self.track.album.replace(artists=[albumartist])
        self.check(self.track.replace(album=album))

    def test_multiple_album_artist_musicbrainz_id(self):
        self.tags["musicbrainz-albumartistid"].append("id")
        self.check(self.track)

    def test_stream_organization_track_name(self):
        del self.tags["title"]
        self.tags["organization"] = ["organization"]
        self.check(self.track.replace(name="organization"))

    def test_multiple_organization_track_name(self):
        del self.tags["title"]
        self.tags["organization"] = ["organization1", "organization2"]
        self.check(self.track.replace(name="organization1; organization2"))

    # TODO: combine all comment types?
    def test_stream_location_track_comment(self):
        del self.tags["comment"]
        self.tags["location"] = ["location"]
        self.check(self.track.replace(comment="location"))

    def test_multiple_location_track_comment(self):
        del self.tags["comment"]
        self.tags["location"] = ["location1", "location2"]
        self.check(self.track.replace(comment="location1; location2"))

    def test_stream_copyright_track_comment(self):
        del self.tags["comment"]
        self.tags["copyright"] = ["copyright"]
        self.check(self.track.replace(comment="copyright"))

    def test_multiple_copyright_track_comment(self):
        del self.tags["comment"]
        self.tags["copyright"] = ["copyright1", "copyright2"]
        self.check(self.track.replace(comment="copyright1; copyright2"))

    def test_sortname(self):
        self.tags["musicbrainz-sortname"] = ["another_sortname"]
        artist = Artist(
            name="artist",
            sortname="another_sortname",
            musicbrainz_id="artistid",
        )
        self.check(self.track.replace(artists=[artist]))

    def test_missing_sortname(self):
        del self.tags["musicbrainz-sortname"]
        artist = Artist(name="artist", sortname=None, musicbrainz_id="artistid")
        self.check(self.track.replace(artists=[artist]))



import logging
import re
import socket

logger = logging.getLogger(__name__)


def try_ipv6_socket():
    """Determine if system really supports IPv6"""
    if not socket.has_ipv6:
        return False
    try:
        socket.socket(socket.AF_INET6).close()
        return True
    except OSError as exc:
        logger.debug(
            f"Platform supports IPv6, but socket creation failed, "
            f"disabling: {exc}"
        )
    return False


#: Boolean value that indicates if creating an IPv6 socket will succeed.
has_ipv6 = try_ipv6_socket()


def format_hostname(hostname):
    """Format hostname for display."""
    if has_ipv6 and re.match(r"\d+.\d+.\d+.\d+", hostname) is not None:
        hostname = f"::ffff:{hostname}"
    return hostname
